// USB controller -> ATARI ver 5.20
// Copyright
//   メインプログラム、コントローラタイプ(TYPE_PS4, TYPE_MDmini, TYPE_XBOXONE, TYPE_XBOX360,TYPE_PCENGNmini,TYPE_NEOGEOmini,TYPE_MOUSE1,TYPE_MOUSE2,TYPE_PS3,TYPE_PS5,TYPE_CYBERSTICK,TYPE_ZUIKI1) :
//     HUYE 2018-2023
//   コントローラタイプ(TYPE_PSC, TYPE_RAP3, TYPE_RAP4, TYPE_SNES, TYPE_RETROFREAK, TYPE_EGRETII) : 
//     たねけん 2020,2022-2023
//   モード(IF_MODE_XE1AP) Special thanks to イカ
//
#include <EEPROM.h>
#include "joyuniversal.h"
#include <max_LCD.h>

#define MAX_JOYSTICK 4
#define AD_CNV_DATA1 64
#define AD_CNV_DATA2 192

#define IF_MODE_ATARI_AB 0
#define IF_MODE_ATARI_BA 1
#define IF_MODE_TWINPLUS1_AB 2
#define IF_MODE_TWINPLUS1_BA 3
#define IF_MODE_CYBER 4
#define IF_MODE_XE1AP 5
#define IF_MODE_JOYDRV 6
#define IF_MODE_JAMMA_MD3B 7
#define IF_MODE_TWINPLUS2_AB 8
#define IF_MODE_TWINPLUS2_BA 9
#define IF_MODE_CPSF    10
#define IF_MODE_MAGICAL 11
#define IF_MODE_MD3B 12
#define IF_MODE_MD6B 13
#define IF_MODE_JAMMA_CPSF 14
#define IF_MODE_JAMMA_MAGICAL 15

#define MODE_CHG_TIME 3000 // 3s
#define RENSHA_SPEED 31 // 31ms
#define SW_CHANGE_MARGIN 1500 //1.5s
#define MD6B_RESETTIME 4 // 4ms
#define OUTLIMIT_TIME 64 // 64ms
#define JOYDRV_WAIT  64 // 64ms

#ifdef dobogusinclude
#include <spi4teensy3.h>
#endif
#include <SPI.h>

byte flg_chg;
byte gpio_sw1,gpio_sw2;
byte modechg_sts[2];
unsigned int usbtask_time;
volatile byte int0_mode, int1_mode;
byte w_int0_mode, w_int1_mode;
volatile byte cyber_speed1, cyber_speed2;
int gpio_cnt;

struct {
  byte data_point;
  byte mask_data;
} directinput_cnvdata[MAX_JOYSTICK][26];

volatile byte md6b_cnt[2];
volatile unsigned int md6b_time[2];
unsigned int mode_time[2];
volatile byte output_time[2];
volatile byte cnv_mode[2];
volatile byte atari_data_SELL[2];
volatile byte atari_data_SELH[2];
volatile byte md6b_data_SELL[2];
volatile byte md6b_data_SELH[2];

volatile byte out_portb_h_atari;
volatile byte out_portb_l_atari;
volatile byte out_portd_h_atari;
volatile byte out_portd_l_atari;
volatile byte out_portb_h_md6b;
volatile byte out_portb_l_md6b;
volatile byte out_portd_h_md6b;
volatile byte out_portd_l_md6b;

volatile byte cyber_data[12];
volatile byte cyber_data_adl = lowByte((unsigned int)cyber_data);
volatile byte cyber_data_adh = highByte((unsigned int)cyber_data);

volatile byte joydrv_rcv_portno;
volatile byte joydrv_rcv_motor1;
volatile byte joydrv_rcv_motor2;
volatile byte joydrv_phase_cnt;
volatile byte joydrv_connect_cnt;
volatile byte joydrv_pointer_adl;
volatile byte joydrv_pointer_adh;

volatile byte joydrv_snddata[MAX_JOYSTICK][16];
volatile byte joydrv_snddata_adl[MAX_JOYSTICK] = {lowByte((unsigned int)joydrv_snddata[0]), lowByte((unsigned int)joydrv_snddata[1]),
                                                  lowByte((unsigned int)joydrv_snddata[2]), lowByte((unsigned int)joydrv_snddata[3])}; 
volatile byte joydrv_snddata_adh[MAX_JOYSTICK] = {highByte((unsigned int)joydrv_snddata[0]), highByte((unsigned int)joydrv_snddata[1]),
                                                  highByte((unsigned int)joydrv_snddata[2]), highByte((unsigned int)joydrv_snddata[3])};
 //  0: X X X X X X SELECT START
 //  1: R4 R3 R2 R1 L4 L3 L2 L1
 //  2: Z' Z Y X D C B A
 //  3: R右 R左 R下 R上 L右 L左 L下 L上
 //  4: L上下アナログ
 //  5: L左右アナログ
 //  6: R上下アナログ
 //  7: R左右アナログ
 //  8: 左アナログボタン1
 //  9: 左アナログボタン2
 // 10: 左アナログボタン3
 // 11: 左アナログボタン4
 // 12: 右アナログボタン1
 // 13: 右アナログボタン2
 // 14: 右アナログボタン3
 // 15: 右アナログボタン4
volatile struct {
  byte motor1;
  byte motor2;
  byte led_r;
  byte led_g;
  byte led_b;
  byte flush_on;
  byte flush_off;
  bool flg_change;
} stick_ctrldata[MAX_JOYSTICK];
byte rensha_time[2][3];

const byte ps_udlr_data[16] PROGMEM = {B11111110, B11110110, B11110111, B11110101, B11111101, B11111001, B11111011, B11111010,
                                       B11111111, B11111111, B11111111, B11111111, B11111111, B11111111, B11111111, B11111111};

const byte psc_udlr_data[16] PROGMEM = {B11111010, B11111110, B11110110, B11111111, B11111011, B11111111, B11110111, B11111111,
                                        B11111001, B11111101, B11110101, B11111111, B11111111, B11111111, B11111111, B11111111};

const char lcd_mode_data[16][11] PROGMEM = {"ATARI_AB  ", "ATARI_BA  ", "TWIN+1C_AB", "TWIN+1C_BA",
                                            "CYBERSTICK", "XE-1AP    ", "JOYDRV    ", "JAMMA_MD3B",
                                            "TWIN+2C_AB", "TWIN+2C_BA", "CPSF      ", "MAGICAL   ",
                                            "MD 3B     ", "MD 6B     ", "JAMMA_CPSF", "JAMMA_MGCL"};

const char* const lcd_mode_adr[16] PROGMEM = {lcd_mode_data[ 0], lcd_mode_data[ 1], lcd_mode_data[ 2], lcd_mode_data[ 3],
                                              lcd_mode_data[ 4], lcd_mode_data[ 5], lcd_mode_data[ 6], lcd_mode_data[ 7],
                                              lcd_mode_data[ 8], lcd_mode_data[ 9], lcd_mode_data[10], lcd_mode_data[11],
                                              lcd_mode_data[12], lcd_mode_data[13], lcd_mode_data[14], lcd_mode_data[15]};

void clr_joydata(byte port, byte mode)
{
  byte i;

  if(mode == 0) {
    for(i=0;i<=3;i++) joydrv_snddata[port][i] = B11111111;
  }
  for(i=4;i<=7;i++) joydrv_snddata[port][i] = B10000000;
  for(i=8;i<=15;i++) joydrv_snddata[port][i] = 0;
}


class JoystickHID : public JOYUniversal {
public:
  JoystickHID(USB *p) :
  JOYUniversal(p) { };

protected: void wrt_BTN_LPLUSU(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][3] &= B11111110;
  else joydrv_snddata[port_no][3] |= B00000001;
}

protected: void wrt_BTN_LPLUSD(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][3] &= B11111101;
  else joydrv_snddata[port_no][3] |= B00000010;
}

protected: void wrt_BTN_LPLUSL(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][3] &= B11111011;
  else joydrv_snddata[port_no][3] |= B00000100;
}

protected: void wrt_BTN_LPLUSR(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][3] &= B11110111;
  else joydrv_snddata[port_no][3] |= B00001000;
}

protected: void wrt_BTN_A(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][2] &= B11111110;
  else joydrv_snddata[port_no][2] |= B00000001;
}

protected: void wrt_BTN_B(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][2] &= B11111101;
  else joydrv_snddata[port_no][2] |= B00000010;
}

protected: void wrt_BTN_C(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][2] &= B11111011;
  else joydrv_snddata[port_no][2] |= B00000100;
}

protected: void wrt_BTN_X(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][2] &= B11101111;
  else joydrv_snddata[port_no][2] |= B00010000;
}

protected: void wrt_BTN_Y(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][2] &= B11011111;
  else joydrv_snddata[port_no][2] |= B00100000;
}

protected: void wrt_BTN_Z(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][2] &= B10111111;
  else joydrv_snddata[port_no][2] |= B01000000;
}

protected: void wrt_BTN_L1(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][1] &= B11111110;
  else joydrv_snddata[port_no][1] |= B00000001;
}

protected: void wrt_BTN_L2(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][1] &= B11111101;
  else joydrv_snddata[port_no][1] |= B00000010;
}

protected: void wrt_BTN_L3(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][1] &= B11111011;
  else joydrv_snddata[port_no][1] |= B00000100;
}

protected: void wrt_BTN_R1(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][1] &= B11101111;
  else joydrv_snddata[port_no][1] |= B00010000;
}

protected: void wrt_BTN_R2(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][1] &= B11011111;
  else joydrv_snddata[port_no][1] |= B00100000;
}

protected: void wrt_BTN_R3(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][1] &= B10111111;
  else joydrv_snddata[port_no][1] |= B01000000;
}

protected: void wrt_BTN_START(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][0] &= B11111110;
  else joydrv_snddata[port_no][0] |= B00000001;
}

protected: void wrt_BTN_SELECT(byte port_no, byte flg)
{
  if(flg) joydrv_snddata[port_no][0] &= B11111101;
  else joydrv_snddata[port_no][0] |= B00000010;
}

protected:
  virtual void ParseHIDData(JOYUniversal *hid, bool is_rpt_id, uint8_t len, uint8_t *buf)
  {
    byte i,j,k;
    byte port_no;
    byte d_pointer;
    uint8_t w_buf[64];

    port_no=byte(PORT_NO);

/*
    if(len > 0) {
      Serial.print(VID, HEX);
      Serial.print(F(":"));
      Serial.print(PID, HEX);
      Serial.println(F(":"));
      for(i=0;i<len;i++) Serial.println(buf[i], HEX);
    }
*/
    // コントローラ別処理
    switch(JOYTYPE) {
      case TYPE_MDmini: // MDmini
        joydrv_snddata[port_no][5] = byte(buf[3]); // 左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[4]); // 上下アナログ

        wrt_BTN_A(port_no, buf[5]&0x0040); // Aボタン
        wrt_BTN_B(port_no, buf[5]&0x0020); // Bボタン
        wrt_BTN_C(port_no, buf[6]&0x0002); // Cボタン
        wrt_BTN_X(port_no, buf[5]&0x0080); // Xボタン
        wrt_BTN_Y(port_no, buf[5]&0x0010); // Yボタン
        wrt_BTN_Z(port_no, buf[6]&0x0001); // Zボタン
        wrt_BTN_START(port_no, buf[6]&0x0020); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[6]&0x0010); // SELECT(MODE)ボタン
        break;

      case TYPE_PS3: // PS3 (DUAL SHOCK3)
        wrt_BTN_LPLUSU(port_no, buf[2]&0x0010); // 上
        wrt_BTN_LPLUSD(port_no, buf[2]&0x0040); // 下
        wrt_BTN_LPLUSL(port_no, buf[2]&0x0080); // 左
        wrt_BTN_LPLUSR(port_no, buf[2]&0x0020); // 右
        wrt_BTN_A(port_no, buf[3]&0x0040); // Aボタン (cross)
        wrt_BTN_B(port_no, buf[3]&0x0020); // Bボタン (circle)
        wrt_BTN_R1(port_no, buf[3]&0x0008); // R1ボタン
        wrt_BTN_R2(port_no, buf[3]&0x0002); // R2ボタン
        wrt_BTN_R3(port_no, buf[2]&0x0004); // R3ボタン
        wrt_BTN_X(port_no, buf[3]&0x0080); // Xボタン (square)
        wrt_BTN_Y(port_no, buf[3]&0x0010); // Yボタン (triangle)
        wrt_BTN_L1(port_no, buf[3]&0x0004); // L1ボタン
        wrt_BTN_L2(port_no, buf[3]&0x0001); // L2ボタン
        wrt_BTN_L3(port_no, buf[2]&0x0002); // L3ボタン
        wrt_BTN_START(port_no, buf[2]&0x0008); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[2]&0x0001); // SELECTボタン

        joydrv_snddata[port_no][5] = byte(buf[6]); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[7]); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[8]); // R左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[9]); // R上下アナログ
        joydrv_snddata[port_no][8] = byte(buf[18]); // 左アナログボタン
        joydrv_snddata[port_no][12] = byte(buf[19]); // 右アナログボタン

        if(stick_ctrldata[port_no].flg_change) {
          for(i = 0; i < PS3_REPORT_BUFFER_SIZE; i++)
            w_buf[i] = pgm_read_byte(&ps3_report_buffer[i]);

          if(stick_ctrldata[port_no].motor1 != 0) {
            w_buf[3] = 0xfe;
            w_buf[4] = stick_ctrldata[port_no].motor1;
          }
          else {
            w_buf[3] = w_buf[4] = 0;
          }

          if(stick_ctrldata[port_no].motor2 != 0) {
            w_buf[1] = 0xfe;
            w_buf[2] = stick_ctrldata[port_no].motor2;
          }
          else {
            w_buf[1] = w_buf[2] = 0;
          }
          SndRpt_ps3(PS3_REPORT_BUFFER_SIZE, w_buf);
        }
        break;

      case TYPE_MYFLASHF300_DI: // PS3 (MAYFLASH ARCADE STICK F300(DINPUT/PS3,XINPUT/NS))
        wrt_BTN_LPLUSU(port_no, (buf[4]==0)); // 上
        wrt_BTN_LPLUSD(port_no, (buf[4]==255)); // 下

        wrt_BTN_LPLUSL(port_no, (buf[3]==0)); // 左
        wrt_BTN_LPLUSR(port_no, (buf[3]==255)); // 右

        wrt_BTN_A(port_no, buf[0]&0x0002); // Aボタン (cross)
        wrt_BTN_B(port_no, buf[0]&0x0004); // Bボタン (circle)
        wrt_BTN_X(port_no, buf[0]&0x0001); // Xボタン (square)
        wrt_BTN_Y(port_no, buf[0]&0x0008); // Yボタン (triangle)
        wrt_BTN_L1(port_no, buf[0]&0x0010); // L1ボタン (LB)
        wrt_BTN_R1(port_no, buf[0]&0x0020); // R1ボタン (RB)
        wrt_BTN_L2(port_no, buf[0]&0x0040); // L2ボタン (LT)
        wrt_BTN_R2(port_no, buf[0]&0x0080); // R2ボタン (RT)
        wrt_BTN_START(port_no, buf[1]&0x0012); // STARTボタン (SHARE,BACK/VIEW)
        wrt_BTN_SELECT(port_no, buf[1]&0x0001); // SELECTボタン (HOME)
        break;

      case TYPE_MYFLASHF300_XB: // PS3 (MAYFLASH ARCADE STICK F300(XBOX))
        wrt_BTN_LPLUSU(port_no, (buf[4]==0)); // 上
        wrt_BTN_LPLUSD(port_no, (buf[4]==255)); // 下

        wrt_BTN_LPLUSL(port_no, (buf[3]==0)); // 左
        wrt_BTN_LPLUSR(port_no, (buf[3]==255)); // 右

        wrt_BTN_A(port_no, buf[0]&0x0001); // Aボタン (cross)
        wrt_BTN_B(port_no, buf[0]&0x0002); // Bボタン (circle)
        wrt_BTN_X(port_no, buf[0]&0x0008); // Xボタン (square)
        wrt_BTN_Y(port_no, buf[0]&0x0010); // Yボタン (triangle)
        wrt_BTN_L1(port_no, buf[0]&0x0040); // L1ボタン (LB)
        wrt_BTN_R1(port_no, buf[0]&0x0080); // R1ボタン (RB)
        wrt_BTN_L2(port_no, buf[1]&0x0001); // L2ボタン (LT)
        wrt_BTN_R2(port_no, buf[1]&0x0002); // R2ボタン (RT)
        wrt_BTN_START(port_no, buf[1]&0x0018); // STARTボタン (SHARE,BACK/VIEW)
        wrt_BTN_SELECT(port_no, buf[1]&0x0004); // SELECTボタン (HOME)
        break;

      case TYPE_RAP3: // RAP (PS3mode)
        // [[ HORI Real Arcade Pro.V HAYABUSA(PS3 MODE) ------------------------------------------------------
        joydrv_snddata[port_no][3] = pgm_read_byte_near(&ps_udlr_data[byte(buf[2])&B00001111]);

        wrt_BTN_A(port_no, buf[0]&0x0002); // Aボタン (cross)
        wrt_BTN_B(port_no, buf[0]&0x0004); // Bボタン (circle)
        wrt_BTN_R1(port_no, buf[0]&0x0020); // R1ボタン
        wrt_BTN_R2(port_no, buf[0]&0x0080); // R2ボタン
        wrt_BTN_R3(port_no, buf[1]&0x0008); // R3ボタン
        wrt_BTN_X(port_no, buf[0]&0x0001); // Xボタン (square)
        wrt_BTN_Y(port_no, buf[0]&0x0008); // Yボタン (triangle)
        wrt_BTN_L1(port_no, buf[0]&0x0010); // L1ボタン
        wrt_BTN_L2(port_no, buf[0]&0x0040); // L2ボタン
        wrt_BTN_L3(port_no, buf[1]&0x0004); // L3ボタン
        wrt_BTN_START(port_no, buf[1]&0x0002); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[1]&0x0001); // SELECTボタン
        break;

      case TYPE_PSC:
        // [[ Playstation Classic Controller Parser ----------------------------------------------------------
        joydrv_snddata[port_no][3] = pgm_read_byte_near(&psc_udlr_data[(byte(buf[1])>>2)&B00001111]);

        wrt_BTN_A(port_no, buf[0]&0x0004); // Aボタン (cross)
        wrt_BTN_B(port_no, buf[0]&0x0002); // Bボタン (circle)
        wrt_BTN_R1(port_no, buf[0]&0x0080); // R1ボタン
        wrt_BTN_R2(port_no, buf[0]&0x0020); // R2ボタン
        wrt_BTN_X(port_no, buf[0]&0x0008); // Xボタン (square)
        wrt_BTN_Y(port_no, buf[0]&0x0001); // Yボタン (triangle)
        wrt_BTN_L1(port_no, buf[0]&0x0040); // L1ボタン
        wrt_BTN_L2(port_no, buf[0]&0x0010); // L2ボタン
        wrt_BTN_START(port_no, buf[1]&0x0002); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[1]&0x0001); // SELECTボタン
        break;

      case TYPE_SNES:
        // [[ iBUFFALO SNES CLASSIC USB GAMEPAD Parser -------------------------------------------------------
        joydrv_snddata[port_no][5] = byte(buf[0]); // 左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[1]); // 上下アナログ

        wrt_BTN_B(port_no, buf[2]&0x0001); // Aボタン
        wrt_BTN_A(port_no, buf[2]&0x0002); // Bボタン
        wrt_BTN_X(port_no, buf[2]&0x0004); // Xボタン
        wrt_BTN_Y(port_no, buf[2]&0x0008); // Yボタン
        wrt_BTN_L1(port_no, buf[2]&0x0010); // L1ボタン
        wrt_BTN_R1(port_no, buf[2]&0x0020); // R1ボタン
        wrt_BTN_SELECT(port_no, buf[2]&0x0040); // SELECTボタン
        wrt_BTN_START(port_no, buf[2]&0x0080); // STARTボタン
        break;

      case TYPE_RETROFREAK:
        // レトロフリーク コントローラアダプター -------------------------------------------------------------
        joydrv_snddata[port_no][5] = byte(buf[1]); // 左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[2]); // 上下アナログ

        switch(byte(buf[0])) {
          case 1: // NESは未確認
          case 4: // PCE2b PCE 6bは未確認
          case 5: // FC
            wrt_BTN_A(port_no, buf[3]&0x0001); // Aボタン
            wrt_BTN_B(port_no, buf[3]&0x0002); // Bボタン
            wrt_BTN_START(port_no, buf[3]&0x0004); // STARTボタン
            wrt_BTN_SELECT(port_no, buf[3]&0x0008); // SELECTボタン
    
            break;
          case 2: // SFC
            wrt_BTN_B(port_no, buf[3]&0x0001); // Bボタン
            wrt_BTN_A(port_no, buf[3]&0x0002); // Aボタン
            wrt_BTN_Y(port_no, buf[3]&0x0004); // Yボタン
            wrt_BTN_X(port_no, buf[3]&0x0008); // Xボタン
            wrt_BTN_L1(port_no, buf[3]&0x0010); // L1ボタン
            wrt_BTN_R1(port_no, buf[3]&0x0020); // R1ボタン
            wrt_BTN_SELECT(port_no, buf[3]&0x0040); // SELECTボタン
            wrt_BTN_START(port_no, buf[3]&0x0080); // STARTボタン
    
            break;
          case 3: // MegaDrive 6bは未確認
            wrt_BTN_A(port_no, buf[3]&0x0001); // Aボタン
            wrt_BTN_B(port_no, buf[3]&0x0002); // Bボタン
            wrt_BTN_C(port_no, buf[3]&0x0004); // Cボタン
            wrt_BTN_X(port_no, buf[3]&0x0008); // Xボタン
            wrt_BTN_Y(port_no, buf[3]&0x0010); // Yボタン
            wrt_BTN_Z(port_no, buf[3]&0x0020); // Zボタン
            wrt_BTN_START(port_no, buf[3]&0x0040); // STARTボタン
            wrt_BTN_SELECT(port_no, buf[3]&0x0080); // MODEボタン
    
            break;
        }
        break;

      case TYPE_XBOXONE: // XBOX ONE
        if(buf[0] != 0x20) break;

        joydrv_snddata[port_no][5] = byte(buf[11]+128); // L左右アナログ
        joydrv_snddata[port_no][4] = byte((buf[13]+128)^B11111111); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[15]+128); // R左右アナログ
        joydrv_snddata[port_no][6] = byte((buf[17]+128)^B11111111); // R上下アナログ
        joydrv_snddata[port_no][8]  = byte((((uint16_t)buf[7] << 8) | buf[6])/4); // 左アナログボタン
        joydrv_snddata[port_no][12] = byte((((uint16_t)buf[9] << 8) | buf[8])/4); // 右アナログボタン

        joydrv_snddata[port_no][3] = byte((buf[5]^B00001111)|B11110000); // 十字キー

        wrt_BTN_A(port_no, buf[4]&0x0010); // Aボタン
        wrt_BTN_B(port_no, buf[4]&0x0020); // Bボタン
        wrt_BTN_R1(port_no, buf[5]&0x0020); // R1ボタン
        wrt_BTN_R2(port_no, (joydrv_snddata[port_no][12]>127)); // R2ボタン
        wrt_BTN_R3(port_no, buf[5]&0x0080); // R3ボタン
        wrt_BTN_X(port_no, buf[4]&0x0040); // Xボタン
        wrt_BTN_Y(port_no, buf[4]&0x0080); // Yボタン
        wrt_BTN_L1(port_no, buf[5]&0x0010); // L1ボタン
        wrt_BTN_L2(port_no, (joydrv_snddata[port_no][8]>127)); // L2ボタン
        wrt_BTN_L3(port_no, buf[5]&0x0040); // L3ボタン
        wrt_BTN_START(port_no, buf[4]&0x0004); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[4]&0x0008); // SELECTボタン
        
        if(stick_ctrldata[port_no].flg_change) {
          memset(w_buf, 0, sizeof(w_buf));
          w_buf[0]  = w_buf[3]  = 0x09;
          w_buf[1]  = w_buf[4]  = w_buf[11] = 0x00;
          w_buf[2]  = cmdCounter++;
          w_buf[5]  = 0x0F;
          w_buf[6]  = w_buf[8]  = stick_ctrldata[port_no].motor1;
          w_buf[7]  = w_buf[9]  = stick_ctrldata[port_no].motor2;
          w_buf[10] = w_buf[12] = 0xFF;
          SndRpt(13, w_buf);
        }
        break;

      case TYPE_XBOX360: // XBOX 360
        if(buf[0] != 0x00 || buf[1] != 0x14) break;

        joydrv_snddata[port_no][5] = byte(buf[7]+128); // L左右アナログ
        joydrv_snddata[port_no][4] = byte((buf[9]+128)^B11111111); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[11]+128); // R左右アナログ
        joydrv_snddata[port_no][6] = byte((buf[13]+128)^B11111111); // R上下アナログ
        joydrv_snddata[port_no][8]  = byte(buf[4]); // 左アナログボタン
        joydrv_snddata[port_no][12] = byte(buf[5]); // 右アナログボタン

        joydrv_snddata[port_no][3] = byte((buf[2]^B00001111)|B11110000); // 十字キー

        wrt_BTN_A(port_no, buf[3]&0x0010); // Aボタン
        wrt_BTN_B(port_no, buf[3]&0x0020); // Bボタン
        wrt_BTN_R1(port_no, buf[3]&0x0002); // R1ボタン
        wrt_BTN_R2(port_no, (joydrv_snddata[port_no][12]>127)); // R2ボタン
        wrt_BTN_R3(port_no, buf[2]&0x0080); // R3ボタン
        wrt_BTN_X(port_no, buf[3]&0x0040); // Xボタン
        wrt_BTN_Y(port_no, buf[3]&0x0080); // Yボタン
        wrt_BTN_L1(port_no, buf[3]&0x0001); // L1ボタン
        wrt_BTN_L2(port_no, (joydrv_snddata[port_no][8]>127)); // L2ボタン
        wrt_BTN_L3(port_no, buf[2]&0x0040); // L3ボタン
        wrt_BTN_START(port_no, buf[2]&0x0010); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[2]&0x0020); // SELECTボタン
        
        if(stick_ctrldata[port_no].flg_change) {
          memset(w_buf, 0, sizeof(w_buf));
          w_buf[0]  = w_buf[2]  = w_buf[5]  = w_buf[6]  = w_buf[7]  = 0x00;
          w_buf[1]  = 0x08;
          w_buf[3]  = stick_ctrldata[port_no].motor2;
          w_buf[4]  = stick_ctrldata[port_no].motor1;
          SndRpt(8, w_buf);
        }
        break;

      case TYPE_PCENGNmini: // PCエンジンmini
        joydrv_snddata[port_no][3] = pgm_read_byte_near(&ps_udlr_data[(byte(buf[2]))&B00001111]);

        joydrv_snddata[port_no][4] = byte(buf[3]); // L上下アナログ
        joydrv_snddata[port_no][5] = byte(buf[4]); // L左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[5]); // R上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[6]); // R左右アナログ

        wrt_BTN_A(port_no, buf[0]&0x0002); // Aボタン
        wrt_BTN_B(port_no, buf[0]&0x0004); // Bボタン
        wrt_BTN_START(port_no, buf[1]&0x0002); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[1]&0x0001); // SELECTボタン
        break;

      case TYPE_NEOGEOmini: // NEOGEOmini
        joydrv_snddata[port_no][3] = pgm_read_byte_near(&ps_udlr_data[(byte(buf[3]))&B00001111]);

        joydrv_snddata[port_no][4] = byte(buf[4]); // L上下アナログ
        joydrv_snddata[port_no][5] = byte(buf[5]); // L左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[6]); // R上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[7]); // R左右アナログ

        wrt_BTN_A(port_no, buf[1]&0x0002); // Aボタン
        wrt_BTN_B(port_no, buf[1]&0x0010); // Bボタン
        wrt_BTN_X(port_no, buf[1]&0x0001); // Xボタン
        wrt_BTN_Y(port_no, buf[1]&0x0008); // Yボタン
        wrt_BTN_START(port_no, buf[2]&0x0008); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[2]&0x0004); // SELECTボタン
        break;

      case TYPE_MOUSE4: // マウス(Type4 アナログ1.5バイト)
        i = byte(buf[1]-128);
        j = byte((((buf[2]>>4)&0x0f)|((buf[3]<<4)&0xf0))-128);
        k = byte(128-buf[4]);
        goto COM_MOUSE;
      case TYPE_MOUSE2: // マウス(Type2 アナログ1バイト)
        i = byte(buf[1]-128);
        j = byte(buf[2]-128);
        k = byte(128-buf[3]);
        goto COM_MOUSE;
      case TYPE_MOUSE1: // マウス(Type1 アナログ2バイト)
        i = byte(buf[1]-128);
        j = byte(buf[3]-128);
        k = byte(128-buf[5]);
       COM_MOUSE:
        joydrv_snddata[port_no][5] = i; // L左右アナログ
        joydrv_snddata[port_no][4] = j; // L上下アナログ
        joydrv_snddata[port_no][6] = k; // R上下アナログ(ホイール)

        wrt_BTN_A(port_no, buf[0]&0x0001); // Aボタン(左クリック)
        wrt_BTN_B(port_no, buf[0]&0x0002); // Bボタン(右クリック)
        wrt_BTN_C(port_no, buf[0]&0x0004); // Cボタン(ホイールボタン)
        wrt_BTN_X(port_no, buf[0]&0x0008); // Xボタン(戻るボタン)
        wrt_BTN_Y(port_no, buf[0]&0x0010); // Yボタン(進むボタン)
//        else {
//          joydrv_snddata[port_no][4] = joydrv_snddata[port_no][5] = joydrv_snddata[port_no][6] = joydrv_snddata[port_no][7] = B10000000;
//        }
        break;

      case TYPE_MOUSE3: // マウス(イーグレットⅡミニ)
        if(buf[2] < 32 || buf[2] >=224) buf[2] = (buf[2]<<2);
        else buf[2] = (buf[2] < 128 ? 127 : 128);
        if(buf[3] < 32 || buf[3] >=224) buf[3] = (buf[3]<<2);
        else buf[3] = (buf[3] < 128 ? 127 : 128);

        joydrv_snddata[port_no][5] = byte(buf[2]-128); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[3]-128); // L上下アナログ
        joydrv_snddata[port_no][6] = byte(128-buf[4]); // R上下アナログ(ホイール)

        wrt_BTN_A(port_no, buf[0]&0x0080); // Aボタン(左大ボタン)
        wrt_BTN_B(port_no, buf[1]&0x0001); // Bボタン(右大ボタン)
        wrt_BTN_X(port_no, buf[0]&0x0002); // Xボタン(真ん中左ボタン)
        wrt_BTN_Y(port_no, buf[0]&0x0010); // Yボタン(真ん中中央ボタン)
        wrt_BTN_Z(port_no, buf[0]&0x0004); // Zボタン(真ん中右ボタン)
        break;

      case TYPE_PS5: // PS5
        if(buf[0] == 0x01) d_pointer = 0;
        else if(buf[0] == 0x31) {
          if (len < 3) break;
          d_pointer = 2;
        }
        else break;

        joydrv_snddata[port_no][5] = byte(buf[d_pointer+1]); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[d_pointer+2]); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[d_pointer+3]); // R左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[d_pointer+4]); // R上下アナログ
        joydrv_snddata[port_no][8]  = byte(buf[d_pointer+5]); // 左アナログボタン
        joydrv_snddata[port_no][12] = byte(buf[d_pointer+6]); // 右アナログボタン

        joydrv_snddata[port_no][3] = pgm_read_byte_near(&ps_udlr_data[byte(buf[d_pointer+8])&B00001111]);

        wrt_BTN_A(port_no, buf[d_pointer+8]&0x0020); // Aボタン (cross)
        wrt_BTN_B(port_no, buf[d_pointer+8]&0x0040); // Bボタン (circle)
        wrt_BTN_R1(port_no, buf[d_pointer+9]&0x0002); // R1ボタン
        wrt_BTN_R2(port_no, buf[d_pointer+9]&0x0008); // R2ボタン
        wrt_BTN_R3(port_no, buf[d_pointer+9]&0x0080); // R3ボタン
        wrt_BTN_X(port_no, buf[d_pointer+8]&0x0010); // Xボタン (square)
        wrt_BTN_Y(port_no, buf[d_pointer+8]&0x0080); // Yボタン (triangle)
        wrt_BTN_L1(port_no, buf[d_pointer+9]&0x0001); // L1ボタン
        wrt_BTN_L2(port_no, buf[d_pointer+9]&0x0004); // L2ボタン
        wrt_BTN_L3(port_no, buf[d_pointer+9]&0x0040); // L3ボタン
        wrt_BTN_START(port_no, ((buf[d_pointer+9]&0x0020) || (buf[d_pointer+10]&0x0001))); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[d_pointer+9]&0x0010); // SELECTボタン

        if(stick_ctrldata[port_no].flg_change) {
          memset(w_buf, 0, sizeof(w_buf));
          w_buf[0]  = 0x02;
          w_buf[1]  = 0xFF;
          w_buf[2]  = 0xF7;
          w_buf[3]  = stick_ctrldata[port_no].motor1;
          w_buf[4]  = stick_ctrldata[port_no].motor2;
          w_buf[9]  = w_buf[11]  = w_buf[12]  = w_buf[13]  = w_buf[22]  = w_buf[23]  = w_buf[24]  = w_buf[44]  = 0;
          w_buf[39]  = 3;
          w_buf[42]  = 1;
          w_buf[45]  = stick_ctrldata[port_no].led_r;
          w_buf[46]  = stick_ctrldata[port_no].led_g;
          w_buf[47]  = stick_ctrldata[port_no].led_b;
          SndRpt(sizeof(w_buf), w_buf);
        }
        break;

      case TYPE_EGRETII: // EGRET II Mini Pad
        joydrv_snddata[port_no][5] = byte(buf[2]); // 左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[3]); // 上下アナログ

        wrt_BTN_A(port_no, buf[0]&0x0010); // Aボタン
        wrt_BTN_B(port_no, buf[0]&0x0004); // Bボタン
        wrt_BTN_C(port_no, buf[0]&0x0002); // Cボタン
        wrt_BTN_X(port_no, buf[1]&0x0001); // Xボタン
        wrt_BTN_Y(port_no, buf[0]&0x0008); // Yボタン
        wrt_BTN_Z(port_no, buf[0]&0x0001); // Zボタン
        wrt_BTN_START(port_no, buf[0]&0x0040); // STARTボタン
        wrt_BTN_SELECT(port_no, ((buf[0]&0x0080) || (buf[1]&0x0002))); // SELECTボタン,HOMEボタン
        break;

      case TYPE_ELECOM1: // JC-U4013SBK(DirectInput)
        joydrv_snddata[port_no][5] = byte(buf[3]); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[4]); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[5]); // R左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[6]); // R上下アナログ
        joydrv_snddata[port_no][8]  = byte(buf[17]); // 左アナログボタン
        joydrv_snddata[port_no][12] = byte(buf[18]); // 右アナログボタン

        joydrv_snddata[port_no][3] = pgm_read_byte_near(&ps_udlr_data[byte(buf[2])&B00001111]);

        wrt_BTN_A(port_no, buf[0]&0x0004); // Aボタン
        wrt_BTN_B(port_no, buf[0]&0x0008); // Bボタン
        wrt_BTN_X(port_no, buf[0]&0x0001); // Xボタン
        wrt_BTN_Y(port_no, buf[0]&0x0002); // Yボタン
        wrt_BTN_L1(port_no, buf[0]&0x0010); // L1ボタン
        wrt_BTN_R1(port_no, buf[0]&0x0020); // R1ボタン
        wrt_BTN_L2(port_no, buf[0]&0x0040); // L2ボタン
        wrt_BTN_R2(port_no, buf[0]&0x0080); // R2ボタン
        wrt_BTN_L3(port_no, buf[1]&0x0001); // L3ボタン
        wrt_BTN_R3(port_no, buf[1]&0x0002); // R3ボタン
        wrt_BTN_SELECT(port_no, ((buf[1]&0x0004) || (buf[1]&0x0010))); // BACKボタン,GUIDEボタン
        wrt_BTN_START(port_no, buf[1]&0x0008); // STARTボタン
        break;

      case TYPE_ELECOM2: // JC-U4013SBK(XInput)
        joydrv_snddata[port_no][5] = byte(128+buf[7]); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(127-buf[9]); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(128+buf[11]); // R左右アナログ
        joydrv_snddata[port_no][6] = byte(127-buf[13]); // R上下アナログ
        joydrv_snddata[port_no][8]  = byte(buf[4]); // 左アナログボタン
        joydrv_snddata[port_no][12] = byte(buf[5]); // 右アナログボタン

        joydrv_snddata[port_no][3] = (buf[2]^B00001111)|B11110000;

        wrt_BTN_A(port_no, buf[3]&0x0010); // Aボタン
        wrt_BTN_B(port_no, buf[3]&0x0020); // Bボタン
        wrt_BTN_X(port_no, buf[3]&0x0040); // Xボタン
        wrt_BTN_Y(port_no, buf[3]&0x0080); // Yボタン
        wrt_BTN_L1(port_no, buf[3]&0x0001); // L1ボタン
        wrt_BTN_R1(port_no, buf[3]&0x0002); // R1ボタン
        wrt_BTN_L2(port_no, buf[4]>0x007F); // L2ボタン
        wrt_BTN_R2(port_no, (buf[5]>0x007f)); // R2ボタン
        wrt_BTN_L3(port_no, buf[2]&0x0040); // L3ボタン
        wrt_BTN_R3(port_no, buf[2]&0x0080); // R3ボタン
        wrt_BTN_SELECT(port_no, ((buf[2]&0x0020) || (buf[3]&0x0004))); // BACKボタン,GUIDEボタン
        wrt_BTN_START(port_no, buf[2]&0x0010); // STARTボタン
        break;

      case TYPE_ELECOM3: // SYSGRATION ELECOM JC-U3312
        joydrv_snddata[port_no][5] = byte(buf[0]); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[1]); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[2]); // R左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[3]); // R上下アナログ
        joydrv_snddata[port_no][8]  = byte(buf[16]); // L1アナログボタン
        joydrv_snddata[port_no][12] = byte(buf[17]); // R1アナログボタン
        joydrv_snddata[port_no][9]  = byte(buf[18]); // L2アナログボタン
        joydrv_snddata[port_no][13] = byte(buf[19]); // R2アナログボタン
        wrt_BTN_LPLUSU(port_no, buf[8]); // POV上
        wrt_BTN_LPLUSD(port_no, buf[10]); // POV下
        wrt_BTN_LPLUSL(port_no, buf[11]); // POV左
        wrt_BTN_LPLUSR(port_no, buf[9]); // POV右
        wrt_BTN_A(port_no, buf[14]); // Aボタン(cross:3)
        wrt_BTN_B(port_no, buf[13]); // Bボタン(circle:4)
        wrt_BTN_X(port_no, buf[15]); // Xボタン(square:1)
        wrt_BTN_Y(port_no, buf[12]); // Yボタン(triangle:2)
        wrt_BTN_L1(port_no, buf[5]&0x0001); // L1ボタン(5)
        wrt_BTN_R1(port_no, buf[5]&0x0002); // R1ボタン(6)
        wrt_BTN_L2(port_no, buf[5]&0x0004); // L2ボタン(7)
        wrt_BTN_R2(port_no, buf[5]&0x0008); // R2ボタン(8)
        wrt_BTN_L3(port_no, buf[5]&0x0040); // L3ボタン(11)
        wrt_BTN_R3(port_no, buf[5]&0x0080); // R3ボタン(12)
        wrt_BTN_SELECT(port_no, buf[5]&0x0010); // SELECTボタン(9)
        wrt_BTN_START(port_no, buf[5]&0x0020); // STARTボタン(10)
        break;

      case TYPE_ELECOM4: // SYSGRATION ELECOM JC-U2912TBK tane
        joydrv_snddata[port_no][5] = byte(buf[2]); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[3]); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[5]); // R左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[4]); // R上下アナログ

        joydrv_snddata[port_no][3] = pgm_read_byte_near(&ps_udlr_data[byte(buf[1]>>4)&B00001111]);
        
        wrt_BTN_A(port_no, buf[0]&0x0004); // Aボタン(cross:3)
        wrt_BTN_B(port_no, buf[0]&0x0008); // Bボタン(circle:4)
        wrt_BTN_X(port_no, buf[0]&0x0001); // Xボタン(square:1)
        wrt_BTN_Y(port_no, buf[0]&0x0002); // Yボタン(triangle:2)
        wrt_BTN_L1(port_no, buf[0]&0x0010); // L1ボタン(5)
        wrt_BTN_R1(port_no, buf[0]&0x0020); // R1ボタン(6)
        wrt_BTN_L2(port_no, buf[0]&0x0040); // L2ボタン(7)
        wrt_BTN_R2(port_no, buf[0]&0x0080); // R2ボタン(8)
        wrt_BTN_L3(port_no, buf[1]&0x0001); // L3ボタン(9)
        wrt_BTN_R3(port_no, buf[1]&0x0002); // R3ボタン(10)
        wrt_BTN_SELECT(port_no, buf[1]&0x0004); // SELECTボタン(11)
        wrt_BTN_START(port_no, buf[1]&0x0008); // STARTボタン(12)
        break;

      case TYPE_CYBERSTICK: // CYBER STICK
        joydrv_snddata[port_no][4] = byte(buf[6]^B11111111); // L上下アナログ
        joydrv_snddata[port_no][5] = byte(buf[5]); // L左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[4]); // R上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[3]); // R左右アナログ

        wrt_BTN_R1(port_no, buf[0]&0x0001); // Aボタン(R1)
        wrt_BTN_R2(port_no, buf[0]&0x0002); // Bボタン(R2)
        wrt_BTN_L1(port_no, buf[0]&0x0004); // Cボタン(L1)
        wrt_BTN_L2(port_no, buf[0]&0x0008); // Dボタン(L2)
        wrt_BTN_X(port_no, buf[0]&0x0010); // E1ボタン(X)
        wrt_BTN_Y(port_no, buf[0]&0x0020); // E2ボタン(Y)
        wrt_BTN_START(port_no, buf[0]&0x0080); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[0]&0x0040); // SELECTボタン
        break;

      case TYPE_SANWA1: // JY-P35U
        joydrv_snddata[port_no][5] = byte(buf[0]); // 左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[1]); // 上下アナログ

        wrt_BTN_A(port_no, buf[2]&0x0008); // Aボタン
        wrt_BTN_B(port_no, buf[2]&0x0010); // Bボタン
        wrt_BTN_C(port_no, buf[2]&0x0020); // Cボタン
        wrt_BTN_X(port_no, buf[2]&0x0001); // Xボタン
        wrt_BTN_Y(port_no, buf[2]&0x0002); // Yボタン
        wrt_BTN_Z(port_no, buf[2]&0x0004); // Zボタン
        wrt_BTN_L1(port_no, buf[2]&0x0040); // L1ボタン
        wrt_BTN_R1(port_no, buf[2]&0x0080); // R1ボタン
        break;

      case TYPE_ZUIKI1: // X68000Z
        joydrv_snddata[port_no][5] = byte(buf[2]); // 左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[3]); // 上下アナログ

        wrt_BTN_A(port_no, buf[0]&0x0002); // Aボタン(表記はb)
        wrt_BTN_B(port_no, buf[0]&0x0004); // Bボタン(表記はa)
        wrt_BTN_L1(port_no, buf[0]&0x0010); // L1ボタン
        wrt_BTN_R1(port_no, buf[0]&0x0020); // R1ボタン
        wrt_BTN_START(port_no, buf[1]&0x0002); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[1]&0x0001); // SELECTボタン
        break;

      case TYPE_8BITDO_TYPE1_DI:  // 8BIT DO Direct Input
        joydrv_snddata[port_no][5] = byte(buf[4]); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[5]); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[6]); // R左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[7]); // R上下アナログ

        joydrv_snddata[port_no][3] = pgm_read_byte_near(&ps_udlr_data[byte(buf[3])]);

        wrt_BTN_A(port_no, buf[1]&0x0001); // Aボタン
        wrt_BTN_B(port_no, buf[1]&0x0002); // Bボタン
        wrt_BTN_X(port_no, buf[1]&0x0008); // Xボタン
        wrt_BTN_Y(port_no, buf[1]&0x0010); // Yボタン
        wrt_BTN_L1(port_no, buf[1]&0x0040); // L1ボタン
        wrt_BTN_L2(port_no, buf[2]&0x0001); // L2ボタン
        wrt_BTN_L3(port_no, buf[2]&0x0020); // L3ボタン
        wrt_BTN_R1(port_no, buf[1]&0x0080); // R1ボタン
        wrt_BTN_R2(port_no, buf[2]&0x0002); // R2ボタン
        wrt_BTN_R3(port_no, buf[2]&0x0040); // R3ボタン
        wrt_BTN_START(port_no, buf[2]&0x0018); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[2]&0x0004); // SELECTボタン
        break;

      case TYPE_PS4: // PS4
        if(buf[0] == 0x01) d_pointer = 0;
        else if(buf[0] == 0x11) {
          if (len < 4) break;
          d_pointer = 2;
        }
        else break;

        joydrv_snddata[port_no][5] = byte(buf[d_pointer+1]); // L左右アナログ
        joydrv_snddata[port_no][4] = byte(buf[d_pointer+2]); // L上下アナログ
        joydrv_snddata[port_no][7] = byte(buf[d_pointer+3]); // R左右アナログ
        joydrv_snddata[port_no][6] = byte(buf[d_pointer+4]); // R上下アナログ
        joydrv_snddata[port_no][8]  = byte(buf[d_pointer+8]); // 左アナログボタン
        joydrv_snddata[port_no][12] = byte(buf[d_pointer+9]); // 右アナログボタン

        joydrv_snddata[port_no][3] = pgm_read_byte_near(&ps_udlr_data[byte(buf[d_pointer+5])&B00001111]);

        wrt_BTN_A(port_no, buf[d_pointer+5]&0x0020); // Aボタン (cross)
        wrt_BTN_B(port_no, buf[d_pointer+5]&0x0040); // Bボタン (circle)
        wrt_BTN_R1(port_no, buf[d_pointer+6]&0x0002); // R1ボタン
        wrt_BTN_R2(port_no, buf[d_pointer+6]&0x0008); // R2ボタン
        wrt_BTN_R3(port_no, buf[d_pointer+6]&0x0080); // R3ボタン
        wrt_BTN_X(port_no, buf[d_pointer+5]&0x0010); // Xボタン (square)
        wrt_BTN_Y(port_no, buf[d_pointer+5]&0x0080); // Yボタン (triangle)
        wrt_BTN_L1(port_no, buf[d_pointer+6]&0x0001); // L1ボタン
        wrt_BTN_L2(port_no, buf[d_pointer+6]&0x0004); // L2ボタン
        wrt_BTN_L3(port_no, buf[d_pointer+6]&0x0040); // L3ボタン
        wrt_BTN_START(port_no, ((buf[d_pointer+6]&0x0020) || (buf[d_pointer+7]&0x0001))); // STARTボタン
        wrt_BTN_SELECT(port_no, buf[d_pointer+6]&0x0010); // SELECTボタン

        if(stick_ctrldata[port_no].flg_change) {
          memset(w_buf, 0, sizeof(w_buf));
          w_buf[0]  = 0x05;
          w_buf[1]  = 0xFF;
          w_buf[4]  = stick_ctrldata[port_no].motor2;
          w_buf[5]  = stick_ctrldata[port_no].motor1;
          w_buf[6]  = stick_ctrldata[port_no].led_r;
          w_buf[7]  = stick_ctrldata[port_no].led_g;
          w_buf[8]  = stick_ctrldata[port_no].led_b;
          w_buf[9]  = stick_ctrldata[port_no].flush_on;
          w_buf[10] = stick_ctrldata[port_no].flush_off;
          SndRpt(sizeof(w_buf), w_buf);
        }
        break;

      case TYPE_DIRECTINPUT: // DirectInput コントローラ

        break;
    }
    stick_ctrldata[port_no].flg_change = false;
  }
};

USB Usb;
//volatile JOYHub Hub(&Usb);
volatile JoystickHID Hid0(&Usb);
volatile JoystickHID Hid1(&Usb);
volatile JoystickHID Hid2(&Usb);
volatile JoystickHID Hid3(&Usb);
volatile JoystickHID Hid4(&Usb);
Max_LCD lcd(&Usb);

void setup() {
  byte i;

  int0_mode = int1_mode = 0;
  cyber_speed1 = cyber_speed2 = 127;
  Serial.begin(115200);
  while (!Serial);

  atari_data_SELL[0] = atari_data_SELH[0] = atari_data_SELL[1] = atari_data_SELH[1] = B11111111;
  out_portb_h_atari = out_portb_l_atari = out_portd_h_atari = out_portd_l_atari = out_portb_h_md6b = out_portb_l_md6b = out_portd_h_md6b = out_portd_l_md6b = B11111111;

  for(i=0;i<MAX_JOYSTICK;i++) {
    clr_joydata(i, 0);
    stick_ctrldata[i].motor1=0;
    stick_ctrldata[i].motor2=0;
    stick_ctrldata[i].led_r=0;
    stick_ctrldata[i].led_g=0;
    stick_ctrldata[i].led_b=0;
    stick_ctrldata[i].flush_on=0;
    stick_ctrldata[i].flush_off=0;
    stick_ctrldata[i].flg_change=false;
  }

  if (Usb.Init() == -1) {
//    Serial.print(F("\r\nOSC did not start"));
    while (1);
  }

  lcd.begin(16,2);
  lcd.noBlink();
  lcd.noCursor();
  lcd.clear();

  modechg_sts[0] = EEPROM.read(2);
  if(modechg_sts[0] > 1) modechg_sts[0] = 0;
  if(modechg_sts[0] == 0) {
    cnv_mode[0] = EEPROM.read(0);
    if(cnv_mode[0] > 15) cnv_mode[0] = IF_MODE_ATARI_AB;
  }
  modechg_sts[1] = EEPROM.read(3);
  if(modechg_sts[1] > 1) modechg_sts[1] = 0;
  if(modechg_sts[1] == 0) {
    cnv_mode[1] = EEPROM.read(1);
    if(cnv_mode[1] > 15  || cnv_mode[1] == IF_MODE_JOYDRV || cnv_mode[1] == IF_MODE_TWINPLUS2_AB || cnv_mode[1] == IF_MODE_JAMMA_CPSF || cnv_mode[1] == IF_MODE_JAMMA_MD3B || cnv_mode[1] == IF_MODE_TWINPLUS2_BA || cnv_mode[1] == IF_MODE_JAMMA_MAGICAL)
      cnv_mode[1] = IF_MODE_ATARI_AB;
  }
  gpio_sw1 = gpio_sw2 = Usb.gpioRd();
  gpio_sw1 &= 0x0f;
  gpio_sw2 = (gpio_sw2>>4)&0x0f;
  gpio_cnt = -1;
  flg_chg = 1;  // モード切替状態として、モードのLED初期表示を行う

  DDRB |= B00000011;
  PORTB |= B00000011;
  DDRD |= B11110000;
  PORTD |= B11110000;

  wait(1000);

  TCCR2A = B00000010;  // CTCモード
  TCCR2B = B00000111;  // 1024分周
  OCR2A = 16;  // 16分周 約1ms

//  TIMSK0 &= B11111110;  // Timer0割り込み禁止
//  TIMSK2 |= B00000010;  // Timer割り込み許可
    /* 時間処理変数の初期化 */

  md6b_time[0] = md6b_time[1] = mode_time[0] = mode_time[1] = 0;
  rensha_time[0][0] = rensha_time[0][1]= rensha_time[0][2] = rensha_time[1][0] = rensha_time[1][1] = rensha_time[1][2] = 254;
  output_time[0] = output_time[1] = 0;

  joydrv_phase_cnt = 0;
  joydrv_connect_cnt = 0;

  usbtask_time = 0;
}

void wait(unsigned int time)
{
  unsigned int i;

  for(i=0;i<time;i++) delayMicroseconds(1000);
}

byte set_cnv_mode(byte port)
{
  byte i,mode;

  DDRB |= B00100000;
//  delay(800);
  mode = ((port == 1 && (cnv_mode[0] == IF_MODE_JOYDRV || cnv_mode[0] == IF_MODE_TWINPLUS2_AB || cnv_mode[0] == IF_MODE_JAMMA_CPSF || cnv_mode[0] == IF_MODE_JAMMA_MD3B || cnv_mode[0] == IF_MODE_TWINPLUS2_BA || cnv_mode[0] == IF_MODE_JAMMA_MAGICAL)) ? cnv_mode[0] : cnv_mode[port]);

  if(modechg_sts[port] == 1) {
    for(i=0;i<5;i++) {
      PORTB |= B00100000; // LED13 ON
      wait(50);
      PORTB &= B11011111; // LED13 OFF
      wait(50);
    }
    wait(450);
  }

  if(mode&0x01) {
    PORTB |= B00100000; // LED13 ON
    wait(1000);
    PORTB &= B11011111; // LED13 OFF
    wait(500);
  }
  i = ((mode>>1)&0x07)+1;
  for(;i>0;i--) {
    PORTB |= B00100000; // LED13 ON
    wait(250);
    PORTB &= B11011111; // LED13 OFF
    if(i != 1) wait(250);
  }

  switch(mode) {
    case IF_MODE_ATARI_AB:
    case IF_MODE_ATARI_BA:
    case IF_MODE_TWINPLUS1_AB:
    case IF_MODE_TWINPLUS1_BA:
    case IF_MODE_CPSF:
    case IF_MODE_MAGICAL:
    case IF_MODE_MD3B:
      if(port == 0) {
        DDRC = B00111111;
        PORTC = B11111111;
        w_int0_mode = 1;
        EICRA = (EICRA&B00001100)|B00000001; // 論理変化
        EIMSK |= B00000001; // 割り込み許可
      }
      else {
        w_int1_mode = 1;
        EICRA = (EICRA&B00000011)|B00000100; // 論理変化
        EIMSK |= B00000010; // 割り込み許可
      }
      break;
    case IF_MODE_JOYDRV:
      w_int0_mode = 4;
      w_int1_mode = 0;
      if(port == 0) {
          DDRC = B00001111;
          PORTC = B11111111;
          EICRA = (EICRA&B00001100)|B00000001; // 論理変化
//        EICRA = (EICRA&B00001100)|B00000010; // 下降
        EIMSK &= B11111101; // 割り込み禁止
        EIMSK |= B00000001; // 割り込み許可
        joydrv_phase_cnt = 0;
        joydrv_connect_cnt = 0;
      }
      break;
    case IF_MODE_CYBER:
    case IF_MODE_XE1AP:
      if(port == 0) {
        DDRC = B00111111;
        PORTC = B11111111;
        w_int0_mode = 3;
        EICRA = (EICRA&B00001100)|B00000010; // 下降
        EIMSK |= B00000001; // 割り込み許可
      }
      else {
        PORTB |= B00000011;
        w_int1_mode = 3;
        EICRA = (EICRA&B00000011)|B00001000; // 下降
        EIMSK |= B00000010; // 割り込み許可
      }
      break;
    case IF_MODE_MD6B:
      md6b_cnt[port] = 0;
      if(port == 0) {
        DDRC = B00111111;
        PORTC = B11111111;
        w_int0_mode = 2;
        EICRA = (EICRA&B00001100)|B00000001; // 論理変化
        EIMSK |= B00000001; // 割り込み許可
      }
      else {
        w_int1_mode = 2;
        EICRA = (EICRA&B00000011)|B00000100; // 論理変化
        EIMSK |= B00000010; // 割り込み許可
      }
      break;
    case IF_MODE_TWINPLUS2_AB:
    case IF_MODE_TWINPLUS2_BA:
    case IF_MODE_JAMMA_CPSF:
    case IF_MODE_JAMMA_MAGICAL:
    case IF_MODE_JAMMA_MD3B:
      w_int0_mode = w_int1_mode = 0;
      EIMSK &= B11111100; // 割り込み禁止
      if(port == 0) {
        DDRC = B00111111;
        PORTC = B11111111;
      }
      break;
  }
  DDRB &= B11011111;
  return mode;
}

void dsplcd_mode(byte port, byte mode)
{
  char work_buf[17];

  lcd.setCursor(0, port);
  work_buf[0] = 'P';
  work_buf[1] = port + 0x31;
  work_buf[2] = '[';
  work_buf[3] = (modechg_sts[port] == 0 ? 'C' : 'S');
  work_buf[4] = ']';
  work_buf[5] = ':';
  strcpy_P(&work_buf[6], pgm_read_word(&(lcd_mode_adr[mode])));
  lcd.print(work_buf);
}

void loop() {
  byte chg_mode;
  byte gpio_work1,gpio_work2;
  byte i;

  if(TIFR2 & B00000010) {
    mode_time[0]++;
    mode_time[1]++;
    md6b_time[0]++;
    md6b_time[1]++;

    if(!(joydrv_snddata[0][2]&0x10)) { // Xボタン
      if(++rensha_time[0][0]>RENSHA_SPEED*2) rensha_time[0][0] = 0;
    }
    else rensha_time[0][0] = 254;

    if(!(joydrv_snddata[0][2]&0x20)) { // Yボタン
      if(++rensha_time[0][1]>RENSHA_SPEED*2) rensha_time[0][1] = 0;
    }
    else rensha_time[0][1] = 254;

    if(((cnv_mode[0] == IF_MODE_MD3B || cnv_mode[0] == IF_MODE_JAMMA_MD3B) && (!(joydrv_snddata[0][1]&0x01) || !(joydrv_snddata[0][1]&0x20))) || !(joydrv_snddata[0][2]&0x40)) { // L1ボタン or R2ボタン or Zボタン
      if(++rensha_time[0][2]>RENSHA_SPEED*2) rensha_time[0][2] = 0;
    }
    else rensha_time[0][2] = 254;

    if(!(joydrv_snddata[1][2]&0x10)) { // Xボタン
      if(++rensha_time[1][0]>RENSHA_SPEED*2) rensha_time[1][0] = 0;
    }
    else rensha_time[1][0] = 254;

    if(!(joydrv_snddata[1][2]&0x20)) { // Yボタン
      if(++rensha_time[1][1]>RENSHA_SPEED*2) rensha_time[1][1] = 0;
    }
    else rensha_time[1][1] = 254;

    if(((cnv_mode[1] == IF_MODE_MD3B || cnv_mode[1] == IF_MODE_JAMMA_MD3B) && (!(joydrv_snddata[1][1]&0x01) || !(joydrv_snddata[1][1]&0x20))) || !(joydrv_snddata[1][2]&0x40)) { // L1ボタン or R2ボタン or Zボタン
      if(++rensha_time[1][2]>RENSHA_SPEED*2) rensha_time[1][2] = 0;
    }
    else rensha_time[1][2] = 254;

    output_time[0]++;
    output_time[1]++;
    if(gpio_cnt >= 0) gpio_cnt++;
    usbtask_time++;
    if(joydrv_phase_cnt>0) {
      if(++joydrv_connect_cnt>JOYDRV_WAIT) {
        joydrv_phase_cnt=0;
        joydrv_connect_cnt=0;
        DDRC = B00001111;
        PORTC = B11111111;
      }
    }
    TIFR2 |= B00000010;
  }

  if(usbtask_time) {
    Usb.Task();
    usbtask_time = 0;
  }

  if(md6b_time[0] >= MD6B_RESETTIME) {
    md6b_cnt[0] = 0;
    md6b_time[0] = 0;
  }   
  if(md6b_time[1] >= MD6B_RESETTIME) {
    md6b_cnt[1] = 0;
    md6b_time[1] = 0;
  }

  for(i=0;i<2;i++) {
    if(!(joydrv_snddata[i][0]&0x01)) { //STARTボタン
      chg_mode = cnv_mode[i];
      if(!(joydrv_snddata[i][2]&0x01)) { // Aボタン
        if(modechg_sts[i] != 0) mode_time[i] = 0;
        else chg_mode = ((joydrv_snddata[i][0]&0x02) ? IF_MODE_ATARI_AB : IF_MODE_ATARI_BA);
      }
      else if(!(joydrv_snddata[i][2]&0x02)) { // Bボタン
        if(modechg_sts[i] != 0) mode_time[i] = 0;
        else chg_mode = ((joydrv_snddata[i][0]&0x02) ? IF_MODE_TWINPLUS1_AB : IF_MODE_TWINPLUS1_BA);
      }
      else if(!(joydrv_snddata[i][2]&0x10)) { // Xボタン
        if(modechg_sts[i] != 0) mode_time[i] = 0;
        else chg_mode = ((joydrv_snddata[i][0]&0x02) ? IF_MODE_CYBER : IF_MODE_XE1AP);
      }
      else if(!(joydrv_snddata[i][2]&0x20)) { // Yボタン
        if(i==1 || modechg_sts[i] != 0) mode_time[i] = 0;
        else chg_mode = ((joydrv_snddata[i][0]&0x02) ? IF_MODE_JOYDRV : IF_MODE_JAMMA_MD3B);
      }
      else if(!(joydrv_snddata[i][3]&0x01) || joydrv_snddata[i][4] < AD_CNV_DATA1) { // 十字キー上 or 左レバー上
        if(i==1 || modechg_sts[i] != 0) mode_time[i] = 0;
        else chg_mode = ((joydrv_snddata[i][0]&0x02) ? IF_MODE_TWINPLUS2_AB : IF_MODE_TWINPLUS2_BA);
      }
      else if(!(joydrv_snddata[i][3]&0x02) || joydrv_snddata[i][4] > AD_CNV_DATA2) { // 十字キー下 or 左レバー下
        if(modechg_sts[i] != 0) mode_time[i] = 0;
        else chg_mode = ((joydrv_snddata[i][0]&0x02) ? IF_MODE_CPSF : IF_MODE_MAGICAL);
      }
      else if(!(joydrv_snddata[i][3]&0x04) || joydrv_snddata[i][5] < AD_CNV_DATA1) { // 十字キー左 or 左レバー左
        if(modechg_sts[i] != 0) mode_time[i] = 0;
        else chg_mode = ((joydrv_snddata[i][0]&0x02) ? IF_MODE_MD3B : IF_MODE_MD6B);
      }
      else if(!(joydrv_snddata[i][3]&0x08) || joydrv_snddata[i][5] > AD_CNV_DATA2) { // 十字キー右 or 左レバー右
        if(i==1 || modechg_sts[i] != 0) mode_time[i] = 0;
        else chg_mode = ((joydrv_snddata[i][0]&0x02) ? IF_MODE_JAMMA_CPSF : IF_MODE_JAMMA_MAGICAL);
      }
      else if(!(joydrv_snddata[i][0]&0x02)) // SELECTボタン
        chg_mode = 255;
      else mode_time[i] = 0;
      if(mode_time[i] >= MODE_CHG_TIME) {
        flg_chg = 1;
        if(chg_mode == 255) {
          modechg_sts[i] ^= B00000001;
          EEPROM.write(i+2, modechg_sts[i]);
        }        
        else {
          cnv_mode[i] = chg_mode;
          EEPROM.write(i, chg_mode);
        }
      }
    }
    else mode_time[i] = 0;
  }

  gpio_work1 = gpio_work2 = Usb.gpioRd();
  gpio_work1 &= 0x0f;
  gpio_work2 = (gpio_work2>>4)&0x0f;

  if(gpio_sw1 != gpio_work1) {
    gpio_sw1 = gpio_work1;
    if(modechg_sts[0] == 1) gpio_cnt = 0;
  }
  if(gpio_sw2 != gpio_work2) {
    gpio_sw2 = gpio_work2;
    if(modechg_sts[1] == 1) gpio_cnt = 0;
  }
  if(gpio_cnt >= SW_CHANGE_MARGIN) {
    gpio_cnt = -1;
    flg_chg = 1;
  }

  if(flg_chg == 1) {
    if(modechg_sts[0] == 1) {
      cnv_mode[0] = (gpio_sw1^B00001111);
      EEPROM.write(0, cnv_mode[0]);
    }
    if(modechg_sts[1] == 1) {
      cnv_mode[1] = (gpio_sw2^B00001111);
      if(cnv_mode[1] == IF_MODE_JOYDRV || cnv_mode[1] == IF_MODE_TWINPLUS2_AB || cnv_mode[1] == IF_MODE_JAMMA_CPSF || cnv_mode[1] == IF_MODE_JAMMA_MD3B || cnv_mode[1] == IF_MODE_TWINPLUS2_BA || cnv_mode[1] == IF_MODE_JAMMA_MAGICAL)
        cnv_mode[1] = IF_MODE_ATARI_AB;
      EEPROM.write(1, cnv_mode[1]);
    }

    int0_mode = int1_mode = 0;
    SPI.end();
    gpio_work1 = set_cnv_mode(0);
    wait(1000);
    gpio_work2 = set_cnv_mode(1);
    SPI.begin();

    dsplcd_mode(0, gpio_work1);
    dsplcd_mode(1, gpio_work2);

    flg_chg = 0;
    mode_time[0] = mode_time[1] = 0; /* モード変更時間の初期化 */ 
//    pre_time = millis();
    int0_mode = w_int0_mode;
    int1_mode = w_int1_mode;
  }

  // ATARI_P1側処理
  switch(cnv_mode[0]) {
    case IF_MODE_ATARI_AB:
    case IF_MODE_ATARI_BA:
      snd_atari(0,cnv_mode[0]);
      break;
    case IF_MODE_TWINPLUS1_AB:
    case IF_MODE_TWINPLUS1_BA:
      snd_twinplus(0,cnv_mode[0]);
      break;
    case IF_MODE_CPSF:
    case IF_MODE_MAGICAL:
      snd_cpsf(0, cnv_mode[0]);
      break;
    case IF_MODE_MD3B:
      snd_md3b(0);
      break;
    case IF_MODE_JAMMA_MD3B:
    case IF_MODE_TWINPLUS2_AB:
    case IF_MODE_TWINPLUS2_BA:
    case IF_MODE_JAMMA_CPSF:
    case IF_MODE_JAMMA_MAGICAL:
      snd_jamma();
      break;
    case IF_MODE_MD6B:
      snd_md6b(0);
      break;
  } // JOYDRVIFは処理なし

  // ATARI_P2側処理
  if(!(cnv_mode[0] == IF_MODE_JOYDRV || cnv_mode[0] == IF_MODE_TWINPLUS2_AB || cnv_mode[0] == IF_MODE_JAMMA_CPSF || cnv_mode[0] == IF_MODE_JAMMA_MD3B || cnv_mode[0] == IF_MODE_TWINPLUS2_BA || cnv_mode[0] == IF_MODE_JAMMA_MAGICAL)) {
    switch(cnv_mode[1]) {
      case IF_MODE_ATARI_AB:
      case IF_MODE_ATARI_BA:
        snd_atari(1,cnv_mode[1]);
        break;
      case IF_MODE_TWINPLUS1_AB:
      case IF_MODE_TWINPLUS1_BA:
        snd_twinplus(1,cnv_mode[1]);
        break;
      case IF_MODE_CPSF:
      case IF_MODE_MAGICAL:
        snd_cpsf(1, cnv_mode[1]);
        break;
      case IF_MODE_MD3B:
        snd_md3b(1);
        break;
      case IF_MODE_MD6B:
        snd_md6b(1);
        break;
    }
  }
}

byte set_lplusad(byte port)
{
  byte work;

  work = (joydrv_snddata[port][3]|B11110000);
  if(joydrv_snddata[port][5] < AD_CNV_DATA1) work &= B11111011; // 左
  else if(joydrv_snddata[port][5] > AD_CNV_DATA2) work &= B11110111; // 右
  if(joydrv_snddata[port][4] < AD_CNV_DATA1) work &= B11111110; // 上
  else if(joydrv_snddata[port][4] > AD_CNV_DATA2) work &= B11111101; // 下

  return work;
}

byte set_rplusad(byte port)
{
  byte work;

  work = ((joydrv_snddata[port][3]>>4)|B11110000);
  if(joydrv_snddata[port][7] < AD_CNV_DATA1) work &= B11111011; // 左
  else if(joydrv_snddata[port][7] > AD_CNV_DATA2) work &= B11110111; // 右
  if(joydrv_snddata[port][6] < AD_CNV_DATA1) work &= B11111110; // 上
  else if(joydrv_snddata[port][6] > AD_CNV_DATA2) work &= B11111101; // 下

  return work;
}

void set_port2data0()
{
  out_portd_h_atari = (atari_data_SELH[1]<<4)|B00001111;
  out_portd_l_atari = (atari_data_SELL[1]<<4)|B00001111;
  out_portb_h_atari = (atari_data_SELH[1]>>4)&B00000011;
  out_portb_l_atari = (atari_data_SELL[1]>>4)&B00000011;
}

void set_port2data1()
{
  asm volatile
  (
    "push r16          \n" // r16退避
    "push r17          \n" // r17退避
    "sbic 9,3          \n" // PIND:bit3 mode1
    "rjmp set_port2data1_10            \n"
    "lds r16,(out_portd_l_atari)   \n"
    "lds r17,(out_portb_l_atari)   \n"
    "rjmp set_port2data1_20            \n"
    "set_port2data1_10:    \n"
    "lds r16,(out_portd_h_atari)   \n"
    "lds r17,(out_portb_h_atari)   \n"
    "set_port2data1_20:    \n"
    "sbrc r16,4        \n"
    "sbi 11,4           \n" // PORTD:4 SET
    "sbrs r16,4        \n"
    "cbi 11,4           \n" // PORTD:4 CLR
    "sbrc r16,5        \n"
    "sbi 11,5           \n" // PORTD:5 SET
    "sbrs r16,5        \n"
    "cbi 11,5           \n" // PORTD:5 CLR
    "sbrc r16,6        \n"
    "sbi 11,6           \n" // PORTD:6 SET
    "sbrs r16,6        \n"
    "cbi 11,6           \n" // PORTD:6 CLR
    "sbrc r16,7        \n"
    "sbi 11,7           \n" // PORTD:7 SET
    "sbrs r16,7        \n"
    "cbi 11,7           \n" // PORTD:7 CLR
    "sbrc r17,0        \n"
    "sbi 5,0           \n" // PORTB:0 SET
    "sbrs r17,0        \n"
    "cbi 5,0           \n" // PORTB:0 CLR
    "sbrc r17,1        \n"
    "sbi 5,1           \n" // PORTB:1 SET
    "sbrs r17,1        \n"
    "cbi 5,1           \n" // PORTB:1 CLR
    "pop r17           \n" // r17復帰
    "pop r16           \n" // r16復帰
  );

//  byte work;
//  work = ((PIND&B00001000) ? out_portd_h_atari : out_portd_l_atari);
//  if(work&B00010000) PORTD |= B00010000;
//  else PORTD &= B11101111;
//  if(work&B00100000) PORTD |= B00100000;
//  else PORTD &= B11011111;
//  if(work&B01000000) PORTD |= B01000000;
//  else PORTD &= B10111111;
//  if(work&B10000000) PORTD |= B10000000;
//  else PORTD &= B01111111;
//  PORTB = (((PIND&B00001000) ? out_portb_h_atari : out_portb_l_atari)|(PORTB&B11111100));
}

void set_port2data2()
{
  asm volatile
  (
    "push r16          \n" // r16退避
    "push r17          \n" // r17退避
    "sbic 9,3          \n" // PIND:bit3
    "rjmp set_port2data2_10            \n"
    "lds r16,(out_portd_l_md6b)   \n"
    "lds r17,(out_portb_l_md6b)   \n"
    "rjmp set_port2data2_20            \n"
    "set_port2data2_10:    \n"
    "lds r16,(out_portd_h_md6b)   \n"
    "lds r17,(out_portb_h_md6b)   \n"
    "set_port2data2_20:    \n"
    "sbrc r16,4        \n"
    "sbi 11,4           \n" // PORTD:4 SET
    "sbrs r16,4        \n"
    "cbi 11,4           \n" // PORTD:4 CLR
    "sbrc r16,5        \n"
    "sbi 11,5           \n" // PORTD:5 SET
    "sbrs r16,5        \n"
    "cbi 11,5           \n" // PORTD:5 CLR
    "sbrc r16,6        \n"
    "sbi 11,6           \n" // PORTD:6 SET
    "sbrs r16,6        \n"
    "cbi 11,6           \n" // PORTD:6 CLR
    "sbrc r16,7        \n"
    "sbi 11,7           \n" // PORTD:7 SET
    "sbrs r16,7        \n"
    "cbi 11,7           \n" // PORTD:7 CLR
    "sbrc r17,0        \n"
    "sbi 5,0           \n" // PORTB:0 SET
    "sbrs r17,0        \n"
    "cbi 5,0           \n" // PORTB:0 CLR
    "sbrc r17,1        \n"
    "sbi 5,1           \n" // PORTB:1 SET
    "sbrs r17,1        \n"
    "cbi 5,1           \n" // PORTB:1 CLR
    "pop r17           \n" // r17復帰
    "pop r16           \n" // r16復帰
  );

//  byte work;
//  work = ((PIND&B00001000) ? out_portd_h_md6b : out_portd_l_md6b);
//  if(work&B00010000) PORTD |= B00010000;
//  else PORTD &= B11101111;
//  if(work&B00100000) PORTD |= B00100000;
//  else PORTD &= B11011111;
//  if(work&B01000000) PORTD |= B01000000;
//  else PORTD &= B10111111;
//  if(work&B10000000) PORTD |= B10000000;
//  else PORTD &= B01111111;
//  PORTB = (((PIND&B00001000) ? out_portb_h_md6b : out_portb_l_md6b)|(PORTB&B11111100));
}

void set_ataridata(byte port, byte mode)
{
  byte atari_work_SELL;
  byte atari_work_SELH;
  byte inf_btn_a,inf_btn_b,inf_btn_ar,inf_btn_br,inf_btn_r1,inf_btn_r2,inf_btn_l1,inf_btn_l2;

  if(mode == IF_MODE_ATARI_AB || mode == IF_MODE_TWINPLUS1_AB || mode == IF_MODE_TWINPLUS2_AB) {
    inf_btn_a  = 0x01;
    inf_btn_ar  = 0;
    inf_btn_b  = 0x02;
    inf_btn_br  = 1;
    inf_btn_r1 = 0x10;
    inf_btn_r2 = 0x20;
    inf_btn_l1 = 0x01;
    inf_btn_l2 = 0x02;
  }
  else {
    inf_btn_a  = 0x02;
    inf_btn_ar  = 1;
    inf_btn_b  = 0x01;
    inf_btn_br  = 0;
    inf_btn_r1 = 0x20;
    inf_btn_r2 = 0x10;
    inf_btn_l1 = 0x02;
    inf_btn_l2 = 0x01;
  }

  atari_work_SELL = set_lplusad(port);
  atari_work_SELH = set_rplusad(port);
  if(!(joydrv_snddata[port][0]&0x01)) // STARTボタン
    atari_work_SELL &= B11110011;
  if(!(joydrv_snddata[port][0]&0x02)) // SELECTボタン
    atari_work_SELL &= B11111100;

  if(!(joydrv_snddata[port][2]&inf_btn_a) || !(joydrv_snddata[port][1]&inf_btn_l1) || rensha_time[port][inf_btn_ar]<RENSHA_SPEED) // 1P Aボタン
    atari_work_SELL &= B11101111;
  if(!(joydrv_snddata[port][1]&inf_btn_r1)) // 2P Aボタン
    atari_work_SELH &= B11101111;
  if(!(joydrv_snddata[port][2]&inf_btn_b) || !(joydrv_snddata[port][1]&inf_btn_l2) || rensha_time[port][inf_btn_br]<RENSHA_SPEED) // 1P Bボタン
    atari_work_SELL &= B11011111;
  if(!(joydrv_snddata[port][1]&inf_btn_r2)) // 2P Bボタン
    atari_work_SELH &= B11011111;
  if(!(joydrv_snddata[port][2]&0x04) || rensha_time[port][2]<RENSHA_SPEED) // 1P Cボタン
    atari_work_SELL &= B11001111;

  if(mode == IF_MODE_ATARI_AB || mode == IF_MODE_ATARI_BA) {
    atari_work_SELL &= atari_work_SELH;
    atari_work_SELH = atari_work_SELL;
  }
  atari_data_SELL[port] = atari_work_SELL;
  atari_data_SELH[port] = atari_work_SELH;
}

void set_cpsfdata(byte port, byte mode)
{
  byte atari_work_SELL;
  byte atari_work_SELH;

  atari_work_SELL = set_lplusad(port);
  atari_work_SELL &= set_rplusad(port);
  atari_work_SELH = ((mode == IF_MODE_CPSF) ? B11111111 : B11111100);

  if(mode == IF_MODE_CPSF) {
    if(!(joydrv_snddata[port][0]&0x01)) // STARTボタン
      atari_work_SELH &= B11011111;
    if(!(joydrv_snddata[port][0]&0x02)) // SELECTボタン
      atari_work_SELH &= B11110111;
  }
  if(!(joydrv_snddata[port][2]&0x01)) // Aボタン
    atari_work_SELL &= ((mode == IF_MODE_CPSF) ? B11101111 : B11011111);
  if(!(joydrv_snddata[port][2]&0x02)) // Bボタン
    atari_work_SELL &= ((mode == IF_MODE_CPSF) ? B11011111 : B11101111);
  if(!(joydrv_snddata[port][1]&0x10) || !(joydrv_snddata[port][1]&0x02) || !(joydrv_snddata[port][2]&0x04)) // R1ボタンorL2ボタンorCボタン
    atari_work_SELH &= ((mode == IF_MODE_CPSF) ? B11111110 : B11110111);
  if(!(joydrv_snddata[port][2]&0x10)) // Xボタン
    atari_work_SELH &= ((mode == IF_MODE_CPSF) ? B11111011 : B11011111);
  if(!(joydrv_snddata[port][2]&0x20)) // Yボタン
    atari_work_SELH &= ((mode == IF_MODE_CPSF) ? B11111101 : B11101111);
  if(!(joydrv_snddata[port][1]&0x01) || !(joydrv_snddata[port][1]&0x20) || !(joydrv_snddata[port][2]&0x40)) // L1ボタンorR2ボタンorZボタン
    atari_work_SELH &= ((mode == IF_MODE_CPSF) ? B11101111 : B11111011);
  atari_data_SELL[port] = atari_work_SELL;
  atari_data_SELH[port] = atari_work_SELH;
}

void set_mddata(byte port,byte mode)
{
  byte atari_work_SELL;
  byte atari_work_SELH;
  byte md6b_work_SELL;
  byte md6b_work_SELH;

  atari_work_SELH = set_lplusad(port);
  atari_work_SELH &= set_rplusad(port);
  atari_work_SELL = (atari_work_SELH&B11110011);
  md6b_work_SELH = B11111111;

  if(!(joydrv_snddata[port][0]&0x01)) // STARTボタン
    atari_work_SELL &= B11011111;
  if(!(joydrv_snddata[port][2]&0x01) || (rensha_time[port][0]<RENSHA_SPEED && (mode == IF_MODE_MD3B || mode == IF_MODE_JAMMA_MD3B))) // Aボタン
    atari_work_SELL &= B11101111;
  if(!(joydrv_snddata[port][2]&0x02) || (rensha_time[port][1]<RENSHA_SPEED && (mode == IF_MODE_MD3B || mode == IF_MODE_JAMMA_MD3B))) // Bボタン
    atari_work_SELH &= B11101111;
  if(!(joydrv_snddata[port][1]&0x10) || !(joydrv_snddata[port][1]&0x02) || !(joydrv_snddata[port][2]&0x04) || (rensha_time[port][2]<RENSHA_SPEED && (mode == IF_MODE_MD3B || mode == IF_MODE_JAMMA_MD3B))) // R1ボタンorL2ボタンorCボタン
    atari_work_SELH &= B11011111;

  if(!(joydrv_snddata[port][0]&0x02)) // SELECTボタン
    md6b_work_SELH &= B11110111;
  if(!(joydrv_snddata[port][2]&0x10)) // Xボタン
    md6b_work_SELH &= B11111011;
  if(!(joydrv_snddata[port][2]&0x20)) // Yボタン
    md6b_work_SELH &= B11111101;
  if(!(joydrv_snddata[port][1]&0x01) || !(joydrv_snddata[port][1]&0x20) || !(joydrv_snddata[port][2]&0x40)) // L1ボタンorR2ボタンorZボタン
    md6b_work_SELH &= B11111110;
  md6b_work_SELL = (atari_work_SELL&B11110000);

  atari_data_SELL[port] = atari_work_SELL;
  atari_data_SELH[port] = atari_work_SELH;
  md6b_data_SELL[port] = md6b_work_SELL;
  md6b_data_SELH[port] = md6b_work_SELH;
}

void snd_atari(byte port,byte mode)
{
  set_ataridata(port, mode);
  if(port == 0) PORTC = ((PIND&B00000100) ? atari_data_SELH[0] : atari_data_SELL[0]);
  else {
    set_port2data0();
    set_port2data1();
  }
}

void snd_twinplus(byte port,byte mode)
{
  set_ataridata(port, mode);
  if(output_time[port] >= OUTLIMIT_TIME) {
    output_time[port] = 0;
    if(port == 0) PORTC = ((PIND&B00000100) ? atari_data_SELH[0] : atari_data_SELL[0]);
    else {
      set_port2data0();
      set_port2data1();
    }
  }
}

void snd_cpsf(byte port, byte mode)
{
  set_cpsfdata(port, mode);
  if(output_time[port] >= OUTLIMIT_TIME) {
    output_time[port] = 0;
    if(port == 0) PORTC = ((PIND&B00000100) ? atari_data_SELH[0] : atari_data_SELL[0]);
    else {
      set_port2data0();
      set_port2data1();
    }
  }
}

void snd_md3b(byte port)
{
  set_mddata(port, IF_MODE_MD3B);
  if(output_time[port] >= OUTLIMIT_TIME) {
    output_time[port] = 0;
    if(port == 0) PORTC = ((PIND&B00000100) ? atari_data_SELH[0] : atari_data_SELL[0]);
    else {
      set_port2data0();
      set_port2data1();
    }
  }
}

void snd_md6b(byte port)
{
  set_mddata(port, IF_MODE_MD6B);
  if(output_time[port] >= OUTLIMIT_TIME) {
    output_time[port] = 0;
    if(port == 0) {
      if(md6b_cnt[0] == 3) PORTC = ((PIND&B00000100) ? md6b_data_SELH[0] : md6b_data_SELL[0]);
      else PORTC = ((PIND&B00000100) ? atari_data_SELH[0] : atari_data_SELL[0]);
    }
    else {
      out_portd_h_md6b = (md6b_data_SELH[1]<<4)|B00001111;
      out_portd_l_md6b = (md6b_data_SELL[1]<<4)|B00001111;
      out_portb_h_md6b = (md6b_data_SELH[1]>>4)&B00000011;
      out_portb_l_md6b = (md6b_data_SELL[1]>>4)&B00000011;
      set_port2data0();
      if(md6b_cnt[1] == 3) set_port2data2();
      else set_port2data1();
    }
  }
}

void snd_jamma()
{
  switch(cnv_mode[0])
  {
    case IF_MODE_JAMMA_CPSF:
      set_cpsfdata(0, IF_MODE_CPSF);
      break;
    case IF_MODE_JAMMA_MAGICAL:
      set_cpsfdata(0, IF_MODE_MAGICAL);
      break;
    case IF_MODE_JAMMA_MD3B:
      set_mddata(0, IF_MODE_JAMMA_MD3B);
      break;
    default: // TWINPLUS2
      set_ataridata(0, cnv_mode[0]);
  }

  out_portd_l_atari = out_portd_h_atari = (atari_data_SELH[0]<<4)|B00001111;
  out_portb_l_atari = out_portb_h_atari = (atari_data_SELH[0]>>4)&B00000011;
  PORTC = atari_data_SELL[0];
  set_port2data1();
}

ISR (INT0_vect,ISR_NAKED)
{
  asm volatile
  (
    "push r16          \n" // r16退避
    "in r16,__SREG__   \n"
    "push r16          \n" // ステータスレジスタ退避

    "lds r16,(int0_mode)   \n"
    "cpi r16,1         \n"
    "brne INT0_MODE1_SKIP           \n"

    "sbic 9,2          \n" // PIND:bit2
    "rjmp INT0_MODE1_10 \n"
    "lds r16,(atari_data_SELL)   \n"
    "out 8,r16         \n" // PORTC
    "rjmp INT0_END          \n"
    "INT0_MODE1_10:    \n"
    "lds r16,(atari_data_SELH)   \n"
    "out 8,r16         \n" // PORTC
    "rjmp INT0_END          \n"

    "INT0_MODE1_SKIP:    \n"
    "cpi r16,2         \n"
    "brne INT0_MODE2_SKIP           \n"

    "lds r16,(md6b_cnt)           \n"
    "sbic 9,2          \n" // PIND:bit2
    "rjmp INT0_MODE2_100           \n"
    "inc r16           \n"
    "sts (md6b_cnt),r16           \n"
    "cpi r16,3         \n"
    "breq INT0_MODE2_10           \n" // md6b_cnt==3の時分岐
    "lds r16,(atari_data_SELL)    \n"
    "out 8,r16         \n" // PORTC
    "rjmp INT0_MODE2_200       \n"
    "INT0_MODE2_10:    \n"
    "lds r16,(md6b_data_SELL)     \n"
    "out 8,r16         \n" // PORTC
    "rjmp INT0_MODE2_200       \n"
    "INT0_MODE2_100:   \n"
    "cpi r16,3         \n"
    "breq INT0_MODE2_110          \n" // md6b_cnt==3の時分岐
    "lds r16,(atari_data_SELH)    \n"
    "out 8,r16         \n" // PORTC
    "rjmp INT0_MODE2_200       \n"
    "INT0_MODE2_110:   \n"
    "lds r16,(md6b_data_SELH)     \n"
    "out 8,r16         \n" // PORTC
    "INT0_MODE2_200:              \n"
    "ldi r16,0         \n"
    "sts (md6b_time),r16         \n"
    "sts (md6b_time+1),r16         \n"
    "rjmp INT0_END            \n"

    "INT0_MODE2_SKIP:    \n"
    "cpi r16,3         \n"
    "breq SND_CYBER0           \n"

    "INT0_MODE3_SKIP:    \n"
    "cpi r16,4         \n"
    "brne INT0_END          \n"
    "jmp SND_JOYDRV          \n"

    "INT0_END:    \n"
    "pop r16           \n" // ステータスレジスタ復帰
    "out __SREG__,r16  \n"
    "pop r16           \n" // r16復帰
    "reti              \n"
  
    // サイバースティックデータ送信
    // CYBER STICK
    //  0 : A+A' B+B' C D
    //  1 : E1 E2 F(START) G(SELECT)
    //  2 : 左上下 上位4bit
    //  3 : 左左右 上位4bit
    //  4 : 右上下 上位4bit
    //  5 : 右左右 上位4bit
    //  6 : 左上下 下位4bit
    //  7 : 左左右 下位4bit
    //  8 : 右上下 下位4bit
    //  9 : 右左右 下位4bit
    // 10 : A B A' B'
    // 11 : 1 1 1 1

    // XE-1AP
    //  0 : E1 E2 F(START) G(SELECT)
    //  1 : A+A' B+B' C D
    //  2 : 左左右 上位4bit
    //  3 : 左上下 上位4bit
    //  4 : 右左右 上位4bit
    //  5 : 右上下 上位4bit
    //  6 : 左左右 下位4bit
    //  7 : 左上下 下位4bit
    //  8 : 右左右 下位4bit
    //  9 : 右上下 下位4bit
    // 10 : E1 E2 E1 E2
    // 11 : 1 1 1 1
    "SND_CYBER0:       \n"
    "push r17          \n" // r17退避
    "push r18          \n" // r18退避
    "push r19          \n" // r19退避
    "push r20          \n" // r20退避
    "push r21          \n" // r21退避
    "push r26       \n"
    "push r27       \n"
    "ldi r16,0xff       \n"
    "sts (cyber_data   ),r16     \n"
    "sts (cyber_data+ 1),r16     \n"
    "sts (cyber_data+ 2),r16     \n"
    "sts (cyber_data+ 3),r16     \n"
    "sts (cyber_data+ 4),r16     \n"
    "sts (cyber_data+ 5),r16     \n"
    "sts (cyber_data+ 6),r16     \n"
    "sts (cyber_data+ 7),r16     \n"
    "sts (cyber_data+ 8),r16     \n"
    "sts (cyber_data+ 9),r16     \n"
    "sts (cyber_data+10),r16     \n"
    "sts (cyber_data+11),r16     \n"

    "lds r16,(cnv_mode)       \n"
    "cpi r16,%0         \n"
    "breq SND_CYBER0_100          \n"
    "rjmp SND_CYBER0_300          \n"

    "SND_CYBER0_100:          \n"
    // 右左右アナログ
    "lds r16,(joydrv_snddata+ 7)       \n"
    "mov r17,r16                        \n"
    "swap r16                           \n"
    "ori r16,0xf0                       \n"
    "ori r17,0xf0                       \n"
    "sts (cyber_data+ 3),r16      \n"
    "sts (cyber_data+ 7),r17      \n"
    // 右上下アナログ
    "lds r16,(joydrv_snddata+ 6)       \n"
    "mov r17,r16                        \n"
    "swap r16                           \n"
    "ori r16,0xf0                       \n"
    "andi r16,0xef                      \n"
    "ori r17,0xf0                       \n"
    "andi r17,0xef                      \n"
    "sts (cyber_data+ 2),r16      \n"
    "sts (cyber_data+ 6),r17      \n"
    // 左左右アナログ
    "lds r16,(joydrv_snddata+ 5)       \n"
    "mov r17,r16                        \n"
    "swap r16                           \n"
    "ori r16,0xf0                       \n"
    "ori r17,0xf0                       \n"
    "sts (cyber_data+ 5),r16      \n"
    "sts (cyber_data+ 9),r17      \n"
    // 左上下アナログ
    "lds r16,(joydrv_snddata+ 4)       \n"
    "mov r17,r16                        \n"
    "swap r16                           \n"
    "ori r16,0xf0                       \n"
    "andi r16,0xef                      \n"
    "ori r17,0xf0                       \n"
    "andi r17,0xef                      \n"
    "sts (cyber_data+ 4),r16      \n"
    "sts (cyber_data+ 8),r17      \n"
    // Bボタン(A')
    "lds r16,(joydrv_snddata+ 2)       \n"
    "sbrc r16,1       \n"
    "rjmp SND_CYBER0_110       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xf7                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "lds r16,(cyber_data+10)       \n"
    "andi r16,0xfd                      \n"
    "sts (cyber_data+10),r16       \n"
    "SND_CYBER0_110:           \n"
    // Aボタン(B')
    "lds r16,(joydrv_snddata+ 2)       \n"
    "sbrc r16,0       \n"
    "rjmp SND_CYBER0_120       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xfb                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "lds r16,(cyber_data+10)       \n"
    "andi r16,0xfe                      \n"
    "sts (cyber_data+10),r16       \n"
    "SND_CYBER0_120:           \n"
    // R1ボタン(A)
    "lds r16,(joydrv_snddata+ 1)       \n"
    "sbrc r16,4       \n"
    "rjmp SND_CYBER0_130       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xf7                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "lds r16,(cyber_data+10)       \n"
    "andi r16,0xf7                      \n"
    "sts (cyber_data+10),r16       \n"
    "SND_CYBER0_130:           \n"
    // R2ボタン(B)
    "lds r16,(joydrv_snddata+ 1)       \n"
    "sbrc r16,5       \n"
    "rjmp SND_CYBER0_140       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xfb                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "lds r16,(cyber_data+10)       \n"
    "andi r16,0xfb                      \n"
    "sts (cyber_data+10),r16       \n"
    "SND_CYBER0_140:           \n"
    // Xボタン(E1)
    "lds r16,(joydrv_snddata+ 2)       \n"
    "sbrc r16,4       \n"
    "rjmp SND_CYBER0_150       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xf7                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "SND_CYBER0_150:           \n"
    // Yボタン(E2)
    "lds r16,(joydrv_snddata+ 2)       \n"
    "sbrc r16,5       \n"
    "rjmp SND_CYBER0_160       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xfb                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "SND_CYBER0_160:           \n"
    // L1ボタン(C)
    "lds r16,(joydrv_snddata+ 1)       \n"
    "sbrc r16,0       \n"
    "rjmp SND_CYBER0_170       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xfd                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "SND_CYBER0_170:           \n"
    // L2ボタン(D)
    "lds r16,(joydrv_snddata+ 1)       \n"
    "sbrc r16,1       \n"
    "rjmp SND_CYBER0_180       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xfe                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "SND_CYBER0_180:           \n"
    // STARTボタン
    "lds r16,(joydrv_snddata+ 0)       \n"
    "sbrc r16,0       \n"
    "rjmp SND_CYBER0_190       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xfd                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "SND_CYBER0_190:           \n"
    // SELECTボタン
    "lds r16,(joydrv_snddata+ 0)       \n"
    "sbrc r16,1       \n"
    "rjmp SND_CYBER0_200       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xfe                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "SND_CYBER0_200:           \n"
    "rjmp SND_CYBER0_500       \n"

    // XE-1AP
    "SND_CYBER0_300:           \n"
    // 右左右アナログ
    "lds r16,(joydrv_snddata+ 7)       \n"
    "mov r17,r16                        \n"
    "swap r16                           \n"
    "ori r16,0xf0                       \n"
    "andi r16,0xef                      \n"
    "ori r17,0xf0                       \n"
    "andi r17,0xef                      \n"
    "sts (cyber_data+ 2),r16      \n"
    "sts (cyber_data+ 6),r17      \n"
    // 右上下アナログ
    "lds r16,(joydrv_snddata+ 6)       \n"
    "mov r17,r16                        \n"
    "swap r16                           \n"
    "ori r16,0xf0                       \n"
    "ori r17,0xf0                       \n"
    "sts (cyber_data+ 3),r16      \n"
    "sts (cyber_data+ 7),r17      \n"
    // 左左右アナログ
    "lds r16,(joydrv_snddata+ 5)       \n"
    "mov r17,r16                        \n"
    "swap r16                           \n"
    "ori r16,0xf0                       \n"
    "andi r16,0xef                      \n"
    "ori r17,0xf0                       \n"
    "andi r17,0xef                      \n"
    "sts (cyber_data+ 4),r16      \n"
    "sts (cyber_data+ 8),r17      \n"
    // 左上下アナログ
    "lds r16,(joydrv_snddata+ 4)       \n"
    "mov r17,r16                        \n"
    "swap r16                           \n"
    "ori r16,0xf0                       \n"
    "ori r17,0xf0                       \n"
    "sts (cyber_data+ 5),r16      \n"
    "sts (cyber_data+ 9),r17      \n"
    // Bボタン(A')
    "lds r16,(joydrv_snddata+ 2)       \n"
    "sbrc r16,1       \n"
    "rjmp SND_CYBER0_310       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xf7                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "lds r16,(cyber_data+10)       \n"
    "andi r16,0xfd                      \n"
    "sts (cyber_data+10),r16       \n"
    "SND_CYBER0_310:           \n"
    // Aボタン(B')
    "lds r16,(joydrv_snddata+ 2)       \n"
    "sbrc r16,0       \n"
    "rjmp SND_CYBER0_320       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xfb                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "lds r16,(cyber_data+10)       \n"
    "andi r16,0xfe                      \n"
    "sts (cyber_data+10),r16       \n"
    "SND_CYBER0_320:           \n"
    // R1ボタン(A)
    "lds r16,(joydrv_snddata+ 1)       \n"
    "sbrc r16,4       \n"
    "rjmp SND_CYBER0_330       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xf7                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "lds r16,(cyber_data+10)       \n"
    "andi r16,0xf7                      \n"
    "sts (cyber_data+10),r16       \n"
    "SND_CYBER0_330:           \n"
    // R2ボタン(B)
    "lds r16,(joydrv_snddata+ 1)       \n"
    "sbrc r16,5       \n"
    "rjmp SND_CYBER0_340       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xfb                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "lds r16,(cyber_data+10)       \n"
    "andi r16,0xfb                      \n"
    "sts (cyber_data+10),r16       \n"
    "SND_CYBER0_340:           \n"
    // Xボタン(E1)
    "lds r16,(joydrv_snddata+ 2)       \n"
    "sbrc r16,4       \n"
    "rjmp SND_CYBER0_350       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xf7                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "SND_CYBER0_350:           \n"
    // Yボタン(E2)
    "lds r16,(joydrv_snddata+ 2)       \n"
    "sbrc r16,5       \n"
    "rjmp SND_CYBER0_360       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xfb                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "SND_CYBER0_360:           \n"
    // L1ボタン(C)
    "lds r16,(joydrv_snddata+ 1)       \n"
    "sbrc r16,0       \n"
    "rjmp SND_CYBER0_370       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xfd                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "SND_CYBER0_370:           \n"
    // L2ボタン(D)
    "lds r16,(joydrv_snddata+ 1)       \n"
    "sbrc r16,1       \n"
    "rjmp SND_CYBER0_380       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xfe                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "SND_CYBER0_380:           \n"
    // STARTボタン
    "lds r16,(joydrv_snddata+ 0)       \n"
    "sbrc r16,0       \n"
    "rjmp SND_CYBER0_390       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xfd                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "SND_CYBER0_390:           \n"
    // SELECTボタン
    "lds r16,(joydrv_snddata+ 0)       \n"
    "sbrc r16,1       \n"
    "rjmp SND_CYBER0_400       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xfe                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "SND_CYBER0_400:           \n"

    "SND_CYBER0_500:           \n"
    "ldi r17,117      \n" // 22us
    "SND_CYBER0_505:           \n"
    "dec r17      \n"
    "brne SND_CYBER0_505      \n"
    "lds r26,cyber_data_adl       \n"
    "lds r27,cyber_data_adh       \n"
    "ldi r16,6      \n"
    "lds r18,cyber_speed1      \n"
    "ldi r19,22      \n"
    "ldi r20,35      \n"
    "ldi r21,21      \n"
    "SND_CYBER0_510:           \n"
    "ld r17,X+       \n"
    "andi r17,0xef         \n" // LH=L,ACK=H
    "out 8,r17         \n" // PORTC
    "ldi r17,117      \n" // 22us
    "SND_CYBER0_520:           \n"
    "dec r17      \n"
    "brne SND_CYBER0_520      \n"
    "cbi 8,5         \n" // PORTC ACK=L
    "sbis 9,2          \n" // PIND:bit2
    "add r19,r20           \n"
    "cpi r18,22           \n"
    "breq SND_CYBER0_526   \n"
    "ldi r21,42      \n"
    "SND_CYBER0_526:           \n"
    "mov r17,r18      \n" // 12us,24us,48us,72us
    "SND_CYBER0_530:           \n"
    "nop      \n"
    "nop      \n"
    "nop      \n"
    "nop      \n"
    "nop      \n"
    "nop      \n"
    "dec r17      \n"
    "brne SND_CYBER0_530      \n"

    "ld r17,X+       \n" // LH=H,ACK=H
    "out 8,r17         \n" // PORTC
    "mov r17,r21      \n" // 4us or 8us
    "SND_CYBER0_540:           \n"
    "dec r17      \n"
    "brne SND_CYBER0_540      \n"
    "cbi 8,5         \n" // PORTC ACK=L
    "mov r17,r18      \n" // 12us,24us,48us,72us
    "SND_CYBER0_550:           \n"
    "nop      \n"
    "nop      \n"
    "nop      \n"
    "nop      \n"
    "nop      \n"
    "nop      \n"
    "dec r17      \n"
    "brne SND_CYBER0_550      \n"

    "dec r16      \n"
    "brne SND_CYBER0_510      \n"
    "sts cyber_speed1,r19      \n"
    "ldi r16,0xff      \n"
    "out 8,r16         \n" // PORTC
    "pop r27       \n"
    "pop r26       \n"
    "pop r21          \n" // r21復帰
    "pop r20          \n" // r20復帰
    "pop r19          \n" // r19復帰
    "pop r18          \n" // r18復帰
    "pop r17          \n" // r17復帰
    "sbi 0x1c,0       \n"
    "rjmp INT0_END            \n"


    "SND_JOYDRV:          \n"
    "push r17          \n" // r17退避
    "push r18          \n" // r18退避
    "push r19          \n" // r19退避
    "push r20          \n" // r19退避
    "push r28          \n" // r28退避
    "push r29          \n" // r29退避
    "push r30          \n" // r30退避
    "push r31          \n" // r31退避

    "lds r16,(joydrv_phase_cnt)   \n"
    "sbic 9,2          \n" // PIND:bit2
    "rjmp SND_JOYDRV_10   \n"
    "sbrs r16,0          \n" // R16:bit0
    "jmp SND_JOYDRV_PAHSEERR   \n"
    "rjmp SND_JOYDRV_20   \n"
    "SND_JOYDRV_10:          \n"
    "sbrc r16,0          \n" // R16:bit0
    "jmp SND_JOYDRV_PAHSEERR   \n"
    "SND_JOYDRV_20:          \n"

    "in r17,6         \n" // PINC
    "swap r17         \n"
    "com r17         \n"
    "andi r17,0x03         \n"
    "mov r18,r17         \n"
    "ldi r19,0xfe         \n"
    "cpi r16,0         \n"
    "breq    SND_JOYDRV_PHASE_0    \n"
    "lds r18,(joydrv_rcv_portno)        \n"
    "lsl r18        \n"
    "lsl r18        \n"
    "or r18,r17        \n"
    "ldi r19,0xfd         \n"
    "cpi r16,1         \n"
    "breq    SND_JOYDRV_PHASE_0    \n"
    "ldi r19,0xfb         \n"
    "cpi r16,2         \n"
    "breq    SND_JOYDRV_PHASE_0    \n"
    "ldi r19,0xf7         \n"
    "cpi r16,3         \n"
    "brne    SND_JOYDRV_PHASE_RS    \n"
    "SND_JOYDRV_PHASE_0:          \n"
    "sts (joydrv_rcv_portno),r18         \n"
    "out 8,r19         \n" // PORTC
    "jmp SND_JOYDRV_100   \n"

    "SND_JOYDRV_PHASE_RS:          \n"
    "cpi r16,4         \n"
    "brne    SND_JOYDRV_PHASE_S0   \n"
    "lds r19,(joydrv_rcv_portno)        \n"
    "lds r30,(joydrv_snddata_adl)         \n"
    "lds r31,(joydrv_snddata_adh)         \n"
    "cpi r19,0         \n"
    "breq    SND_JOYDRV_PHASE_RS_100    \n"
    "lds r30,(joydrv_snddata_adl+1)         \n"
    "lds r31,(joydrv_snddata_adh+1)         \n"
    "cpi r19,1         \n"
    "breq    SND_JOYDRV_PHASE_RS_100    \n"
    "lds r30,(joydrv_snddata_adl+2)         \n"
    "lds r31,(joydrv_snddata_adh+2)         \n"
    "cpi r19,2         \n"
    "breq    SND_JOYDRV_PHASE_RS_100    \n"
    "lds r30,(joydrv_snddata_adl+3)         \n"
    "lds r31,(joydrv_snddata_adh+3)         \n"
    "SND_JOYDRV_PHASE_RS_100:          \n"
    "sts (joydrv_pointer_adl),r30         \n"
    "sts (joydrv_pointer_adh),r31         \n"
    "clr r19         \n"
    "sts (joydrv_rcv_motor1),r19         \n"
    "sts (joydrv_rcv_motor2),r19         \n"
    "ser r19         \n"
    "out 8,r19         \n" // PORTC
    "ldi r19,0x1f         \n"
    "out 7,r19         \n" // DDRC

    "SND_JOYDRV_PHASE_S0:          \n"
    "lds r30,(joydrv_pointer_adl)         \n"
    "lds r31,(joydrv_pointer_adh)         \n"
    "ldd r19,Z+0         \n"
    "cpi r16,4         \n"
    "brne    SND_JOYDRV_PHASE_S0_10    \n"
    "rjmp    SND_JOYDRV_PHASE_S1    \n"
    "SND_JOYDRV_PHASE_S0_10:          \n"
    "cpi r16,5         \n"
    "brne    SND_JOYDRV_PHASE_S0_20    \n"
    "rjmp    SND_JOYDRV_PHASE_S2_00    \n"
    "SND_JOYDRV_PHASE_S0_20:          \n"

    "ldd r19,Z+1         \n"
    "cpi r16,6         \n"
    "brne    SND_JOYDRV_PHASE_S0_30    \n"
    "rjmp    SND_JOYDRV_PHASE_S1    \n"
    "SND_JOYDRV_PHASE_S0_30:          \n"
    "cpi r16,7         \n"
    "brne    SND_JOYDRV_PHASE_S0_40    \n"
    "rjmp    SND_JOYDRV_PHASE_S2_00    \n"
    "SND_JOYDRV_PHASE_S0_40:          \n"

    "ldd r19,Z+2         \n"
    "cpi r16,8         \n"
    "brne    SND_JOYDRV_PHASE_S0_50    \n"
    "rjmp    SND_JOYDRV_PHASE_S1    \n"
    "SND_JOYDRV_PHASE_S0_50:          \n"
    "cpi r16,9         \n"
    "brne    SND_JOYDRV_PHASE_S0_60    \n"
    "rjmp    SND_JOYDRV_PHASE_S2_00    \n"
    "SND_JOYDRV_PHASE_S0_60:          \n"

    "ldd r19,Z+3         \n"
    "cpi r16,10         \n"
    "brne    SND_JOYDRV_PHASE_S0_70    \n"
    "rjmp    SND_JOYDRV_PHASE_S1    \n"
    "SND_JOYDRV_PHASE_S0_70:          \n"
    "cpi r16,11         \n"
    "brne    SND_JOYDRV_PHASE_S0_80    \n"
    "rjmp    SND_JOYDRV_PHASE_S2_00    \n"
    "SND_JOYDRV_PHASE_S0_80:          \n"

    "ldd r19,Z+4         \n"
    "cpi r16,12         \n"
    "breq    SND_JOYDRV_PHASE_S1    \n"
    "cpi r16,13         \n"
    "breq    SND_JOYDRV_PHASE_S2_00    \n"
    "ldd r19,Z+5         \n"
    "cpi r16,14         \n"
    "breq    SND_JOYDRV_PHASE_S1    \n"
    "cpi r16,15         \n"
    "breq    SND_JOYDRV_PHASE_S2_00    \n"
    "ldd r19,Z+6         \n"
    "cpi r16,16         \n"
    "breq    SND_JOYDRV_PHASE_S1    \n"
    "cpi r16,17         \n"
    "breq    SND_JOYDRV_PHASE_S2_00    \n"
    "ldd r19,Z+7         \n"
    "cpi r16,18         \n"
    "breq    SND_JOYDRV_PHASE_S1    \n"
    "cpi r16,19         \n"
    "breq    SND_JOYDRV_PHASE_S2_00    \n"
    "ldd r19,Z+8         \n"
    "cpi r16,20         \n"
    "breq    SND_JOYDRV_PHASE_S1    \n"
    "cpi r16,21         \n"
    "breq    SND_JOYDRV_PHASE_S2_10    \n"
    "ldd r19,Z+9         \n"
    "cpi r16,22         \n"
    "breq    SND_JOYDRV_PHASE_S1    \n"
    "cpi r16,23         \n"
    "breq    SND_JOYDRV_PHASE_S2_10    \n"
    "ldd r19,Z+10         \n"
    "cpi r16,24         \n"
    "breq    SND_JOYDRV_PHASE_S1    \n"
    "cpi r16,25         \n"
    "breq    SND_JOYDRV_PHASE_S2_10    \n"
    "ldd r19,Z+11         \n"
    "cpi r16,26         \n"
    "breq    SND_JOYDRV_PHASE_S1    \n"
    "cpi r16,27         \n"
    "breq    SND_JOYDRV_PHASE_S2_10    \n"
    "ldd r19,Z+12         \n"
    "cpi r16,28         \n"
    "breq    SND_JOYDRV_PHASE_S1    \n"
    "cpi r16,29         \n"
    "breq    SND_JOYDRV_PHASE_S2_10    \n"
    "ldd r19,Z+13         \n"
    "cpi r16,30         \n"
    "breq    SND_JOYDRV_PHASE_S1    \n"
    "cpi r16,31         \n"
    "breq    SND_JOYDRV_PHASE_S2_10    \n"
    "ldd r19,Z+14         \n"
    "cpi r16,32         \n"
    "breq    SND_JOYDRV_PHASE_S1    \n"
    "cpi r16,33         \n"
    "breq    SND_JOYDRV_PHASE_S2_10    \n"
    "ldd r19,Z+15         \n"
    "cpi r16,34         \n"
    "breq    SND_JOYDRV_PHASE_S1    \n"
    "cpi r16,35         \n"
    "breq    SND_JOYDRV_PHASE_S2_10    \n"
    "rjmp SND_JOYDRV_PHASE_END   \n"

    "SND_JOYDRV_PHASE_S1:          \n"
    "swap r19         \n"
    "ori r19,0xf0         \n"
    "andi r19,0xef         \n"
    "out 8,r19         \n" // PORTC
    "rjmp SND_JOYDRV_100   \n"
    "SND_JOYDRV_PHASE_S2_00:          \n"
    "lds r18,(joydrv_rcv_motor1)        \n"
    "lsl r18        \n"
    "sbis 6,5         \n" // PINC bit5
    "sbr r18,0x01         \n"
    "sts (joydrv_rcv_motor1),r18         \n"
    "rjmp SND_JOYDRV_PHASE_S2_20   \n"
    "SND_JOYDRV_PHASE_S2_10:          \n"
    "lds r18,(joydrv_rcv_motor2)        \n"
    "lsl r18        \n"
    "sbis 6,5         \n" // PINC bit5
    "sbr r18,0x01         \n"
    "sts (joydrv_rcv_motor2),r18         \n"
    "SND_JOYDRV_PHASE_S2_20:          \n"
    "ori r19,0xf0         \n"
    "out 8,r19         \n" // PORTC

    "SND_JOYDRV_100:   \n"
    "inc r16   \n"
    "sts (joydrv_phase_cnt),r16   \n"
    "clr r16         \n"
    "sts (joydrv_connect_cnt),r16   \n"
    "rjmp SND_JOYDRV_200   \n"

    "SND_JOYDRV_PHASE_END:          \n"
    "cpi r16,36         \n"
    "breq   SND_JOYDRV_PHASE_END_10     \n"
    "rjmp SND_JOYDRV_PAHSENOM       \n"
    "SND_JOYDRV_PHASE_END_10:          \n"
    "ser r19         \n"
    "out 8,r19         \n" // PORTC
    "ldi r19,0x0f         \n"
    "out 7,r19         \n" // DDRC
    "inc r16   \n"
    "sts (joydrv_phase_cnt),r16   \n"
    "clr r16         \n"
    "sts (joydrv_connect_cnt),r16   \n"

    "lds r17,(joydrv_rcv_motor1)         \n"
    "lds r18,(joydrv_rcv_motor2)         \n"
    "lds r19,(joydrv_rcv_portno)        \n"
    "cpi r19,0         \n"
    "breq    SND_JOYDRV_PHASE_END_200    \n"
    "cpi r19,1         \n"
    "breq    SND_JOYDRV_PHASE_END_300    \n"
    "cpi r19,2         \n"
    "breq    SND_JOYDRV_PHASE_END_400    \n"
    "SND_JOYDRV_PHASE_END_100:          \n"
    "lds r28,(stick_ctrldata+0+24)         \n"
    "lds r29,(stick_ctrldata+1+24)         \n"
    "cp r17,r28         \n"
    "brne    SND_JOYDRV_PHASE_END_110    \n"
    "cp r18,r29         \n"
    "breq    SND_JOYDRV_PHASE_END_120    \n"
    "SND_JOYDRV_PHASE_END_110:          \n"
    "sts (stick_ctrldata+0+24),r17         \n"
    "sts (stick_ctrldata+1+24),r18         \n"
    "ser r20         \n"
    "SND_JOYDRV_PHASE_END_120:          \n"
    "sts (stick_ctrldata+7+24),r20         \n"
    "rjmp SND_JOYDRV_200   \n"

    "SND_JOYDRV_PHASE_END_200:          \n"
    "lds r28,(stick_ctrldata+0)         \n"
    "lds r29,(stick_ctrldata+1)         \n"
    "cp r17,r28         \n"
    "brne    SND_JOYDRV_PHASE_END_210    \n"
    "cp r18,r29         \n"
    "breq    SND_JOYDRV_PHASE_END_220    \n"
    "SND_JOYDRV_PHASE_END_210:          \n"
    "sts (stick_ctrldata+0),r17         \n"
    "sts (stick_ctrldata+1),r18         \n"
    "ser r20         \n"
    "SND_JOYDRV_PHASE_END_220:          \n"
    "sts (stick_ctrldata+7),r20         \n"
    "rjmp SND_JOYDRV_200   \n"

    "SND_JOYDRV_PHASE_END_300:          \n"
    "lds r28,(stick_ctrldata+0+ 8)         \n"
    "lds r29,(stick_ctrldata+1+ 8)         \n"
    "cp r17,r28         \n"
    "brne    SND_JOYDRV_PHASE_END_310    \n"
    "cp r18,r29         \n"
    "breq    SND_JOYDRV_PHASE_END_320    \n"
    "SND_JOYDRV_PHASE_END_310:          \n"
    "sts (stick_ctrldata+0+ 8),r17         \n"
    "sts (stick_ctrldata+1+ 8),r18         \n"
    "ser r20         \n"
    "SND_JOYDRV_PHASE_END_320:          \n"
    "sts (stick_ctrldata+7+ 8),r20         \n"
    "rjmp SND_JOYDRV_200   \n"

    "SND_JOYDRV_PHASE_END_400:          \n"
    "lds r28,(stick_ctrldata+0+16)         \n"
    "lds r29,(stick_ctrldata+1+16)         \n"
    "cp r17,r28         \n"
    "brne    SND_JOYDRV_PHASE_END_410    \n"
    "cp r18,r29         \n"
    "breq    SND_JOYDRV_PHASE_END_420    \n"
    "SND_JOYDRV_PHASE_END_410:          \n"
    "sts (stick_ctrldata+0+16),r17         \n"
    "sts (stick_ctrldata+1+16),r18         \n"
    "ser r20         \n"
    "SND_JOYDRV_PHASE_END_420:          \n"
    "sts (stick_ctrldata+7+16),r20         \n"
    "rjmp SND_JOYDRV_200   \n"

    "SND_JOYDRV_PAHSEERR:   \n"
    "ser r19         \n"
    "out 8,r19         \n" // PORTC
    "ldi r19,0x0f         \n"
    "out 7,r19         \n" // DDRC
    "SND_JOYDRV_PAHSENOM:   \n"
    "clr r16         \n"
    "sts (joydrv_phase_cnt),r16   \n"
    "sts (joydrv_connect_cnt),r16   \n"

    "SND_JOYDRV_200:   \n"
    "sbi 0x1c,0       \n"
    "pop r31          \n" // r31退避
    "pop r30          \n" // r30退避
    "pop r29          \n" // r29退避
    "pop r28          \n" // r28退避
    "pop r20          \n" // r19退避
    "pop r19          \n" // r19退避
    "pop r18          \n" // r18退避
    "pop r17          \n" // r17退避
    "jmp INT0_END            \n"

    : : "M"(IF_MODE_CYBER) :
  );
}

ISR (INT1_vect,ISR_NAKED)
{
  asm volatile
  (
    "push r16          \n" // r16退避
    "in r16,__SREG__   \n"
    "push r16          \n" // ステータスレジスタ退避

    "lds r16,(int1_mode)   \n"
    "cpi r16,1         \n"
    "brne INT1_MODE1_SKIP           \n"

    "push r17          \n" // r17退避
    "sbic 9,3          \n" // PIND:bit3 mode1
    "rjmp INT1_MODE1_10            \n"
    "lds r16,(out_portd_l_atari)   \n"
    "lds r17,(out_portb_l_atari)   \n"
    "rjmp INT1_MODE1_20            \n"
    "INT1_MODE1_10:    \n"
    "lds r16,(out_portd_h_atari)   \n"
    "lds r17,(out_portb_h_atari)   \n"
    "INT1_MODE1_20:    \n"
    "sbrc r16,4        \n"
    "sbi 11,4           \n" // PORTD:4 SET
    "sbrs r16,4        \n"
    "cbi 11,4           \n" // PORTD:4 CLR
    "sbrc r16,5        \n"
    "sbi 11,5           \n" // PORTD:5 SET
    "sbrs r16,5        \n"
    "cbi 11,5           \n" // PORTD:5 CLR
    "sbrc r16,6        \n"
    "sbi 11,6           \n" // PORTD:6 SET
    "sbrs r16,6        \n"
    "cbi 11,6           \n" // PORTD:6 CLR
    "sbrc r16,7        \n"
    "sbi 11,7           \n" // PORTD:7 SET
    "sbrs r16,7        \n"
    "cbi 11,7           \n" // PORTD:7 CLR
    "sbrc r17,0        \n"
    "sbi 5,0           \n" // PORTB:0 SET
    "sbrs r17,0        \n"
    "cbi 5,0           \n" // PORTB:0 CLR
    "sbrc r17,1        \n"
    "sbi 5,1           \n" // PORTB:1 SET
    "sbrs r17,1        \n"
    "cbi 5,1           \n" // PORTB:1 CLR
    "pop r17           \n" // r17復帰
    "rjmp INT1_END            \n"

    "INT1_MODE1_SKIP:    \n"
    "cpi r16,2         \n"
    "brne INT1_MODE2_SKIP           \n"

    "push r17          \n" // r17退避
    "lds r16,(md6b_cnt+1)           \n" // mode2
    "sbic 9,3          \n" // PIND:bit3
    "rjmp INT1_MODE2_100           \n"
    "inc r16           \n"
    "sts (md6b_cnt+1),r16           \n"
    "cpi r16,3         \n"
    "breq INT1_MODE2_10           \n" // md6b_cnt==3の時分岐
    "lds r16,(out_portd_l_atari)   \n"
    "lds r17,(out_portb_l_atari)   \n"
    "rjmp INT1_MODE2_200       \n"
    "INT1_MODE2_10:    \n"
    "lds r16,(out_portd_l_md6b)   \n"
    "lds r17,(out_portb_l_md6b)   \n"
    "rjmp INT1_MODE2_200       \n"
    "INT1_MODE2_100:   \n"
    "cpi r16,3         \n"
    "breq INT1_MODE2_110          \n" // md6b_cnt==3の時分岐
    "lds r16,(out_portd_h_atari)   \n"
    "lds r17,(out_portb_h_atari)   \n"
    "rjmp INT1_MODE2_200       \n"
    "INT1_MODE2_110:   \n"
    "lds r16,(out_portd_h_md6b)   \n"
    "lds r17,(out_portb_h_md6b)   \n"
    "INT1_MODE2_200:              \n"
    "sbrc r16,4        \n"
    "sbi 11,4           \n" // PORTD:4 SET
    "sbrs r16,4        \n"
    "cbi 11,4           \n" // PORTD:4 CLR
    "sbrc r16,5        \n"
    "sbi 11,5           \n" // PORTD:5 SET
    "sbrs r16,5        \n"
    "cbi 11,5           \n" // PORTD:5 CLR
    "sbrc r16,6        \n"
    "sbi 11,6           \n" // PORTD:6 SET
    "sbrs r16,6        \n"
    "cbi 11,6           \n" // PORTD:6 CLR
    "sbrc r16,7        \n"
    "sbi 11,7           \n" // PORTD:7 SET
    "sbrs r16,7        \n"
    "cbi 11,7           \n" // PORTD:7 CLR
    "sbrc r17,0        \n"
    "sbi 5,0           \n" // PORTB:0 SET
    "sbrs r17,0        \n"
    "cbi 5,0           \n" // PORTB:0 CLR
    "sbrc r17,1        \n"
    "sbi 5,1           \n" // PORTB:1 SET
    "sbrs r17,1        \n"
    "cbi 5,1           \n" // PORTB:1 CLR
    "pop r17           \n" // r17復帰
    "ldi r16,0         \n"
    "sts (md6b_time+2),r16         \n"
    "sts (md6b_time+3),r16         \n"
    "rjmp INT1_END            \n"

    "INT1_MODE2_SKIP:    \n"
    "cpi r16,3         \n"
    "breq SND_CYBER1           \n"

    "INT1_END:    \n"
    "pop r16           \n" // ステータスレジスタ復帰
    "out __SREG__,r16  \n"
    "pop r16           \n" // r16復帰
    "reti              \n"

    // サイバースティックデータ送信
    // CYBER STICK
    //  0 : A+A' B+B' C D
    //  1 : E1 E2 F(START) G(SELECT)
    //  2 : 左上下 上位4bit
    //  3 : 左左右 上位4bit
    //  4 : 右上下 上位4bit
    //  5 : 右左右 上位4bit
    //  6 : 左上下 下位4bit
    //  7 : 左左右 下位4bit
    //  8 : 右上下 下位4bit
    //  9 : 右左右 下位4bit
    // 10 : A B A' B'
    // 11 : 1 1 1 1

    // XE-1AP
    //  0 : E1 E2 F(START) G(SELECT)
    //  1 : A+A' B+B' C D
    //  2 : 左左右 上位4bit
    //  3 : 左上下 上位4bit
    //  4 : 右左右 上位4bit
    //  5 : 右上下 上位4bit
    //  6 : 左左右 下位4bit
    //  7 : 左上下 下位4bit
    //  8 : 右左右 下位4bit
    //  9 : 右上下 下位4bit
    // 10 : E1 E2 E1 E2
    // 11 : 1 1 1 1
    "SND_CYBER1:       \n"
    "push r17          \n" // r17退避
    "push r18          \n" // r18退避
    "push r19          \n" // r19退避
    "push r20          \n" // r20退避
    "push r21          \n" // r21退避
    "push r26       \n"
    "push r27       \n"
    "ldi r16,0xff       \n"
    "sts (cyber_data   ),r16     \n"
    "sts (cyber_data+ 1),r16     \n"
    "sts (cyber_data+ 2),r16     \n"
    "sts (cyber_data+ 3),r16     \n"
    "sts (cyber_data+ 4),r16     \n"
    "sts (cyber_data+ 5),r16     \n"
    "sts (cyber_data+ 6),r16     \n"
    "sts (cyber_data+ 7),r16     \n"
    "sts (cyber_data+ 8),r16     \n"
    "sts (cyber_data+ 9),r16     \n"
    "sts (cyber_data+10),r16     \n"
    "sts (cyber_data+11),r16     \n"

    "lds r16,(cnv_mode+1)       \n"
    "cpi r16,%0         \n"
    "breq SND_CYBER1_100          \n"
    "rjmp SND_CYBER1_300          \n"

    "SND_CYBER1_100:          \n"
    // 右左右アナログ
    "lds r16,(joydrv_snddata+ 7+16)       \n"
    "mov r17,r16                        \n"
    "swap r16                           \n"
    "ori r16,0xf0                       \n"
    "ori r17,0xf0                       \n"
    "sts (cyber_data+ 3),r16      \n"
    "sts (cyber_data+ 7),r17      \n"
    // 右上下アナログ
    "lds r16,(joydrv_snddata+ 6+16)       \n"
    "mov r17,r16                        \n"
    "swap r16                           \n"
    "ori r16,0xf0                       \n"
    "andi r16,0xef                      \n"
    "ori r17,0xf0                       \n"
    "andi r17,0xef                      \n"
    "sts (cyber_data+ 2),r16      \n"
    "sts (cyber_data+ 6),r17      \n"
    // 左左右アナログ
    "lds r16,(joydrv_snddata+ 5+16)       \n"
    "mov r17,r16                        \n"
    "swap r16                           \n"
    "ori r16,0xf0                       \n"
    "ori r17,0xf0                       \n"
    "sts (cyber_data+ 5),r16      \n"
    "sts (cyber_data+ 9),r17      \n"
    // 左上下アナログ
    "lds r16,(joydrv_snddata+ 4+16)       \n"
    "mov r17,r16                        \n"
    "swap r16                           \n"
    "ori r16,0xf0                       \n"
    "andi r16,0xef                      \n"
    "ori r17,0xf0                       \n"
    "andi r17,0xef                      \n"
    "sts (cyber_data+ 4),r16      \n"
    "sts (cyber_data+ 8),r17      \n"
    // Bボタン(A')
    "lds r16,(joydrv_snddata+ 2+16)       \n"
    "sbrc r16,1       \n"
    "rjmp SND_CYBER1_110       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xf7                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "lds r16,(cyber_data+10)       \n"
    "andi r16,0xfd                      \n"
    "sts (cyber_data+10),r16       \n"
    "SND_CYBER1_110:           \n"
    // Aボタン(B')
    "lds r16,(joydrv_snddata+ 2+16)       \n"
    "sbrc r16,0       \n"
    "rjmp SND_CYBER1_120       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xfb                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "lds r16,(cyber_data+10)       \n"
    "andi r16,0xfe                      \n"
    "sts (cyber_data+10),r16       \n"
    "SND_CYBER1_120:           \n"
    // R1ボタン(A)
    "lds r16,(joydrv_snddata+ 1+16)       \n"
    "sbrc r16,4       \n"
    "rjmp SND_CYBER1_130       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xf7                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "lds r16,(cyber_data+10)       \n"
    "andi r16,0xf7                      \n"
    "sts (cyber_data+10),r16       \n"
    "SND_CYBER1_130:           \n"
    // R2ボタン(B)
    "lds r16,(joydrv_snddata+ 1+16)       \n"
    "sbrc r16,5       \n"
    "rjmp SND_CYBER1_140       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xfb                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "lds r16,(cyber_data+10)       \n"
    "andi r16,0xfb                      \n"
    "sts (cyber_data+10),r16       \n"
    "SND_CYBER1_140:           \n"
    // Xボタン(E1)
    "lds r16,(joydrv_snddata+ 2+16)       \n"
    "sbrc r16,4       \n"
    "rjmp SND_CYBER1_150       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xf7                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "SND_CYBER1_150:           \n"
    // Yボタン(E2)
    "lds r16,(joydrv_snddata+ 2+16)       \n"
    "sbrc r16,5       \n"
    "rjmp SND_CYBER1_160       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xfb                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "SND_CYBER1_160:           \n"
    // L1ボタン(C)
    "lds r16,(joydrv_snddata+ 1+16)       \n"
    "sbrc r16,0       \n"
    "rjmp SND_CYBER1_170       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xfd                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "SND_CYBER1_170:           \n"
    // L2ボタン(D)
    "lds r16,(joydrv_snddata+ 1+16)       \n"
    "sbrc r16,1       \n"
    "rjmp SND_CYBER1_180       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xfe                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "SND_CYBER1_180:           \n"
    // STARTボタン
    "lds r16,(joydrv_snddata+ 0+16)       \n"
    "sbrc r16,0       \n"
    "rjmp SND_CYBER1_190       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xfd                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "SND_CYBER1_190:           \n"
    // SELECTボタン
    "lds r16,(joydrv_snddata+ 0+16)       \n"
    "sbrc r16,1       \n"
    "rjmp SND_CYBER1_200       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xfe                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "SND_CYBER1_200:           \n"
    "rjmp SND_CYBER1_500       \n"

    // XE-1AP
    "SND_CYBER1_300:           \n"
    // 右左右アナログ
    "lds r16,(joydrv_snddata+ 7+16)       \n"
    "mov r17,r16                        \n"
    "swap r16                           \n"
    "ori r16,0xf0                       \n"
    "andi r16,0xef                      \n"
    "ori r17,0xf0                       \n"
    "andi r17,0xef                      \n"
    "sts (cyber_data+ 2),r16      \n"
    "sts (cyber_data+ 6),r17      \n"
    // 右上下アナログ
    "lds r16,(joydrv_snddata+ 6+16)       \n"
    "mov r17,r16                        \n"
    "swap r16                           \n"
    "ori r16,0xf0                       \n"
    "ori r17,0xf0                       \n"
    "sts (cyber_data+ 3),r16      \n"
    "sts (cyber_data+ 7),r17      \n"
    // 左左右アナログ
    "lds r16,(joydrv_snddata+ 5+16)       \n"
    "mov r17,r16                        \n"
    "swap r16                           \n"
    "ori r16,0xf0                       \n"
    "andi r16,0xef                      \n"
    "ori r17,0xf0                       \n"
    "andi r17,0xef                      \n"
    "sts (cyber_data+ 4),r16      \n"
    "sts (cyber_data+ 8),r17      \n"
    // 左上下アナログ
    "lds r16,(joydrv_snddata+ 4+16)       \n"
    "mov r17,r16                        \n"
    "swap r16                           \n"
    "ori r16,0xf0                       \n"
    "ori r17,0xf0                       \n"
    "sts (cyber_data+ 5),r16      \n"
    "sts (cyber_data+ 9),r17      \n"
    // Bボタン(A')
    "lds r16,(joydrv_snddata+ 2+16)       \n"
    "sbrc r16,1       \n"
    "rjmp SND_CYBER1_310       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xf7                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "lds r16,(cyber_data+10)       \n"
    "andi r16,0xfd                      \n"
    "sts (cyber_data+10),r16       \n"
    "SND_CYBER1_310:           \n"
    // Aボタン(B')
    "lds r16,(joydrv_snddata+ 2+16)       \n"
    "sbrc r16,0       \n"
    "rjmp SND_CYBER1_320       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xfb                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "lds r16,(cyber_data+10)       \n"
    "andi r16,0xfe                      \n"
    "sts (cyber_data+10),r16       \n"
    "SND_CYBER1_320:           \n"
    // R1ボタン(A)
    "lds r16,(joydrv_snddata+ 1+16)       \n"
    "sbrc r16,4       \n"
    "rjmp SND_CYBER1_330       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xf7                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "lds r16,(cyber_data+10)       \n"
    "andi r16,0xf7                      \n"
    "sts (cyber_data+10),r16       \n"
    "SND_CYBER1_330:           \n"
    // R2ボタン(B)
    "lds r16,(joydrv_snddata+ 1+16)       \n"
    "sbrc r16,5       \n"
    "rjmp SND_CYBER1_340       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xfb                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "lds r16,(cyber_data+10)       \n"
    "andi r16,0xfb                      \n"
    "sts (cyber_data+10),r16       \n"
    "SND_CYBER1_340:           \n"
    // Xボタン(E1)
    "lds r16,(joydrv_snddata+ 2+16)       \n"
    "sbrc r16,4       \n"
    "rjmp SND_CYBER1_350       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xf7                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "SND_CYBER1_350:           \n"
    // Yボタン(E2)
    "lds r16,(joydrv_snddata+ 2+16)       \n"
    "sbrc r16,5       \n"
    "rjmp SND_CYBER1_360       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xfb                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "SND_CYBER1_360:           \n"
    // L1ボタン(C)
    "lds r16,(joydrv_snddata+ 1+16)       \n"
    "sbrc r16,0       \n"
    "rjmp SND_CYBER1_370       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xfd                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "SND_CYBER1_370:           \n"
    // L2ボタン(D)
    "lds r16,(joydrv_snddata+ 1+16)       \n"
    "sbrc r16,1       \n"
    "rjmp SND_CYBER1_380       \n"
    "lds r16,(cyber_data+ 1)       \n"
    "andi r16,0xfe                      \n"
    "sts (cyber_data+ 1),r16       \n"
    "SND_CYBER1_380:           \n"
    // STARTボタン
    "lds r16,(joydrv_snddata+ 0+16)       \n"
    "sbrc r16,0       \n"
    "rjmp SND_CYBER1_390       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xfd                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "SND_CYBER1_390:           \n"
    // SELECTボタン
    "lds r16,(joydrv_snddata+ 0+16)       \n"
    "sbrc r16,1       \n"
    "rjmp SND_CYBER1_400       \n"
    "lds r16,(cyber_data+ 0)       \n"
    "andi r16,0xfe                      \n"
    "sts (cyber_data+ 0),r16       \n"
    "SND_CYBER1_400:           \n"

    "SND_CYBER1_500:           \n"
    "ldi r17,117      \n" // 22us
    "SND_CYBER1_505:           \n"
    "dec r17      \n"
    "brne SND_CYBER1_505      \n"
    "lds r26,cyber_data_adl       \n"
    "lds r27,cyber_data_adh       \n"
    "ldi r16,6      \n"
    "lds r18,cyber_speed2      \n"
    "ldi r19,22      \n"
    "ldi r20,35      \n"
    "ldi r21,21      \n"
    "SND_CYBER1_510:           \n"
    "ld r17,X+       \n"
    "sbrc r17,0        \n"
    "sbi 11,4           \n" // PORTD:4 SET
    "sbrs r17,0        \n"
    "cbi 11,4           \n" // PORTD:4 CLR
    "sbrc r17,1        \n"
    "sbi 11,5           \n" // PORTD:5 SET
    "sbrs r17,1        \n"
    "cbi 11,5           \n" // PORTD:5 CLR
    "sbrc r17,2        \n"
    "sbi 11,6           \n" // PORTD:6 SET
    "sbrs r17,2        \n"
    "cbi 11,6           \n" // PORTD:6 CLR
    "sbrc r17,3        \n"
    "sbi 11,7           \n" // PORTD:7 SET
    "sbrs r17,3        \n"
    "cbi 11,7           \n" // PORTD:7 CLR
    "cbi 5,0           \n" // PORTB:0 CLR LH=L
    "sbi 5,1           \n" // PORTB:1 SET ACK=H
    "ldi r17,117      \n" // 22us
    "SND_CYBER1_520:           \n"
    "dec r17      \n"
    "brne SND_CYBER1_520      \n"
    "cbi 5,1           \n" // PORTB:1 CLR ACK=L
    "sbis 9,3          \n" // PIND:bit3
    "add r19,r20           \n"
    "cpi r18,22           \n"
    "breq SND_CYBER1_526   \n"
    "ldi r21,42      \n"
    "SND_CYBER1_526:           \n"
    "mov r17,r18      \n" // 12us,24us,48us,72us
    "SND_CYBER1_530:           \n"
    "nop      \n"
    "nop      \n"
    "nop      \n"
    "nop      \n"
    "nop      \n"
    "nop      \n"
    "dec r17      \n"
    "brne SND_CYBER1_530      \n"

    "ld r17,X+       \n"
    "sbrc r17,0        \n"
    "sbi 11,4           \n" // PORTD:4 SET
    "sbrs r17,0        \n"
    "cbi 11,4           \n" // PORTD:4 CLR
    "sbrc r17,1        \n"
    "sbi 11,5           \n" // PORTD:5 SET
    "sbrs r17,1        \n"
    "cbi 11,5           \n" // PORTD:5 CLR
    "sbrc r17,2        \n"
    "sbi 11,6           \n" // PORTD:6 SET
    "sbrs r17,2        \n"
    "cbi 11,6           \n" // PORTD:6 CLR
    "sbrc r17,3        \n"
    "sbi 11,7           \n" // PORTD:7 SET
    "sbrs r17,3        \n"
    "cbi 11,7           \n" // PORTD:7 CLR
    "sbi 5,0           \n" // PORTB:0 SET LH=H
    "sbi 5,1           \n" // PORTB:1 SET ACK=H
    "mov r17,r21      \n" // 4us or 8us
    "SND_CYBER1_540:           \n"
    "dec r17      \n"
    "brne SND_CYBER1_540      \n"
    "cbi 5,1           \n" // PORTB:1 CLR ACK=L
    "mov r17,r18      \n" // 12us,24us,48us,72us
    "SND_CYBER1_550:           \n"
    "nop      \n"
    "nop      \n"
    "nop      \n"
    "nop      \n"
    "nop      \n"
    "nop      \n"
    "dec r17      \n"
    "brne SND_CYBER1_550      \n"

    "dec r16      \n"
    "breq SND_CYBER1_560      \n"
    "rjmp SND_CYBER1_510      \n"
    "SND_CYBER1_560:           \n"
    "sts cyber_speed2,r19      \n"
    "sbi 11,4           \n" // PORTD:4 SET
    "sbi 11,5           \n" // PORTD:5 SET
    "sbi 11,6           \n" // PORTD:6 SET
    "sbi 11,7           \n" // PORTD:7 SET
    "sbi 5,0           \n" // PORTB:0 SET LH=H
    "sbi 5,1           \n" // PORTB:1 SET ACK=H
    "pop r27       \n"
    "pop r26       \n"
    "pop r21          \n" // r21復帰
    "pop r20          \n" // r20復帰
    "pop r19          \n" // r19復帰
    "pop r18          \n" // r18復帰
    "pop r17          \n" // r17復帰
    "sbi 0x1c,1       \n"
    "rjmp INT1_END            \n"
    : : "M"(IF_MODE_CYBER) :
  );
}
